####################################################################################################
# main/surface_to_image.py
# The code for the function that handles the registration of retinotopy
# By Noah C. Benson

from __future__ import print_function

import numpy                        as     np
import scipy                        as     sp
import nibabel                      as     nib
import nibabel.freesurfer.io        as     fsio
import nibabel.freesurfer.mghformat as     fsmgh
import os, sys, pimms

from ..freesurfer                   import (subject, add_subject_path, find_subject_path)
from ..io                           import (save, load)
from ..mri                          import (is_image, is_image_spec, image_clear, to_image)

info = \
   '''
   Syntax: surface_to_image <subject> <out>
   <subject> must be a valid FreeSurfer subject id (in the $SUBJECTS_DIR directory
     or configured subject paths) or a full path to a FreeSurfer subject or it can
     be a valid HCP subject identifier.
   <out> is the output volume file to write.
   In addition to the subject and the output filename, at least one and at most two
   surface file(s) must be specified. These may be specified using the --lh (or -l)
   and --rh (or -r) options below or without them; if they files are specified
   without the given arguments and the ordering cannot be detected by the number of
   vertices in the file, then the left hemisphere surface is assumed to be
   specified first.
   The command projects the given surface files into the subject's ribbon and
   writes the result out to the given filename. If only one hemisphere's surface
   datafile is given, then only that hemisphere's data is projected to the ribbon.

   The following options may be given:
     * -v|--verbose
       Indicates that detailed output should be printed.
     * -l|--lh=<file>
       -r|--rh=<file>
       These options specify the surface data files that are to be projected to
       the subject's ribbon.
     * -i|--image=<file>
       The given file specifies the image-spec of the output image; all header-data
       are copied from this image and the file's affine and the image shape are used
       to determine the slice prescription of the output image. File may optionally
       be a JSON file containing a mapping with the keys 'affine' and 'image_shape'.
       If no image is given then the subject's 'brain' image is used.
     * -m|--method=<name>
       Specifies the method that should be used. Supported options are linear,
       nearest, and auto (the default). Both linear and nearest methods find the
       position of the voxel-center in the cortical sheet. The linear option linearly
       interpolates between the three vertices that form the triangle faces of the
       prism containing the voxel-center while the nearest option uses the nearest of
       these three vertices. If the auto option is used, then chooses nearest for int
       data and linear for floating-point data.
     * -f|--fill=<value>
       The fill value (default: 0) is inserted in everywhere in the volume that is
       not part of the ribbon.
     * -t|--dtype=<value>
       Specifies that the output data type should be <value>. Currently supported are
       'int' or 'float' (default: 'float').
     * -d|--subjects-dir=<path>
       Specifies that the given path(s) should be added to the subjects directory
       when performing the operation. Note that this may include directories
       separated by colons (:), as with the PATH environment variable.
     * --
       This token, by itself, indicates that the arguments that remain should not
       be processed as flags or options, even if they begin with a -.
   '''
_surface_to_ribbon_parser_instructions = [
    # Flags
    ('h', 'help',         'help',         False),
    ('v', 'verbose',      'verbose',      False),
    # Options             
    ['l', 'lh',           'lh_file',      None],
    ['r', 'rh',           'rh_file',      None],
    ['i', 'image',        'image',        None],
    ['f', 'fill',         'fill',         0],
    ['m', 'method',       'method',       'auto'],
    ['t', 'type',         'dtype',        None],
    ['d', 'subjects-dir', 'subjects_dir', None]]
_surface_to_ribbon_parser = pimms.argv_parser(_surface_to_ribbon_parser_instructions)

def read_surf_file(flnm):
  if flnm.endswith(".mgh") or flnm.endswith(".mgz"):
    data = np.array(fsmgh.load(flnm).dataobj).flatten()
  else:
    data = fsio.read_morph_data(flnm)
  return data

def main(args):
    '''
    surface_to_rubbon.main(args) can be given a list of arguments, such as sys.argv[1:]; these
    arguments may include any options and must include exactly one subject id and one output
    filename. Additionally one or two surface input filenames must be given. The surface files are
    projected into the ribbon and written to the output filename. For more information see the
    string stored in surface_to_image.info.
    '''
    # Parse the arguments
    (args, opts) = _surface_to_ribbon_parser(args)
    # First, help?
    if opts['help']:
        print(info, file=sys.stdout)
        return 1
    # and if we are verbose, lets setup a note function
    verbose = opts['verbose']
    def note(s):
        if verbose: print(s, file=sys.stdout)
        return verbose
    # Add the subjects directory, if there is one
    if 'subjects_dir' in opts and opts['subjects_dir'] is not None:
        add_subject_path(opts['subjects_dir'])
    # figure out our arguments:
    (lhfl, rhfl) = (opts['lh_file'], opts['rh_file'])
    if len(args) == 0:
      raise ValueError('Not enough arguments provided!')
    elif len(args) == 1:
      # must be that the subject is in the env?
      sub = find_subject_path(os.getenv('SUBJECT'))
      outfl = args[0]
    elif len(args) == 2:
      sbpth = find_subject_path(args[0])
      if sbpth is not None:
        sub = sbpth
      else:
        sub = find_subject_path(os.getenv('SUBJECT'))
        if lhfl is not None: rhfl = args[0]
        elif rhfl is not None: lhfl = args[0]
        else: raise ValueError('Given arg is not a subject: %s' % args[0])
      outfl = args[1]
    elif len(args) == 3:
      sbpth0 = find_subject_path(args[0])
      sbpth1 = find_subject_path(args[1])
      if sbpth0 is not None:
        sub = sbpth0
        if lhfl is not None: rhfl = args[1]
        elif rhfl is not None: lhfl = args[1]
        else: raise ValueError('Too many arguments given: %s' % args[1])
      elif sbpth1 is not None:
        sub = sbpth1
        if lhfl is not None: rhfl = args[0]
        elif rhfl is not None: lhfl = args[0]
        else: raise ValueError('Too many arguments given: %s' % args[0])
      else:
        sub = find_subject_path(os.getenv('SUBJECT'))
        if lhfl is not None or rhfl is not None:
          raise ValueError('Too many arguments and no subject given')
        (lhfl, rhfl) = args
      outfl = args[2]
    elif len(args) == 4:
      if lhfl is not None or rhfl is not None:
          raise ValueError('Too many arguments and no subject given')
      subidx = next((i for (i,a) in enumerate(args) if find_subject_path(a) is not None), None)
      if subidx is None: raise ValueError('No subject given')
      sub = find_subject_path(args[subidx])
      del args[subidx]
      (lhfl, rhfl, outfl) = args
    else:
      raise ValueError('Too many arguments provided!')
    if sub is None: raise ValueError('No subject specified or found in $SUBJECT')
    if lhfl is None and rhfl is None: raise ValueError('No surfaces provided')
    sub = subject(sub)
    # check the method
    method = opts['method'].lower()
    if method not in ['linear', 'nearest', 'auto']:
        raise ValueError('Unsupported method: %s' % method)
    # and the datatype
    if opts['dtype'] is None: dtyp = None
    elif opts['dtype'].lower() == 'float': dtyp = np.float
    elif opts['dtype'].lower() == 'int': dtyp = np.int
    else: raise ValueError('Type argument must be float or int')
    if method == 'auto':
      if dtyp is np.float: method = 'linear'
      elif dtyp is np.int: method = 'nearest'
      else: method = 'linear'
    # and the input/sample image
    im = opts['image']
    if im is None:
      im = sub.images['brain']
      try: note('Using template image: %s' % im.get_filename())
      except Exception: pass
    else:
      note('Using template image: %s' % im)
      im = load(im)
    im = to_image(image_clear(im, fill=opts['fill']), dtype=dtyp)
    # Now, load the data:
    note('Reading surfaces...')
    (lhdat, rhdat) = (None, None)
    if lhfl is not None:
        note('   - Reading LH file: %s' % lhfl)
        lhdat = read_surf_file(lhfl)
    if rhfl is not None:
        note('   - Reading RH file: %s' % rhfl)
        rhdat = read_surf_file(rhfl)
    (dat, hemi) = (rhdat, 'rh') if lhdat is None else \
                  (lhdat, 'lh') if rhdat is None else \
                  ((lhdat, rhdat), None)
    # okay, make the volume...
    note('Generating volume...')
    im = sub.cortex_to_image(dat, im, hemi=hemi, method=method, fill=opts['fill'], dtype=dtyp)
    # and write out the file
    note('Exporting volume file: %s' % outfl)
    save(outfl, im)
    note('surface_to_image complete!')
    return 0




####################################################################################################
# command/__init__.py
# The main function, if neuropythy is invoked directly as command.
# By Noah C. Benson

import pyrsistent as _pyr

from . import register_retinotopy as _reg
from . import benson14_retinotopy as _b14
from . import surface_to_image    as _s2i
from . import atlas               as _atl
from . import retinotopy          as _ret

# The commands that can be run by main:
commands = _pyr.m(
    atlas               = _atl.main,
    retinotopy          = _ret.main,
    register_retinotopy = _reg.main,
    benson14_retinotopy = _b14.main,
    surface_to_image    = _s2i.main)

__all__ = ['commands']




####################################################################################################
# neuropythy/datasets/core.py
# Implementation of builtin datasets for Neuropythy
# by Noah C. Benson

import os, six, shutil, tempfile, atexit, pimms
import numpy as np

from ..util import (config, to_credentials, ObjectWithMetaData)
from ..freesurfer import subject as freesurfer_subject

# We declare a configuration variable, data_cache_root -- where to put the data that is downloaded.
# If this is None / unset, then we'll use a temporary directory and auto-delete it on exit.
config.declare_dir('data_cache_root')

@pimms.immutable
class Dataset(ObjectWithMetaData):
    '''
    The Dataset class is a simple immutable class that should be implemented by all neuropythy
    datasets. The design is such that neuropythy.data[name] should always (lazily) yield a Dataset
    object specific to the dataset given by name, if it exists and can be loaded.
    
    One reason to require (by convention) that all datasets are distinct classes is that it should
    thus be easy to evaluate help(ny.data[name]) to see help on the given dataset. If you overload
    this class, be sure to overload the documentation.
    '''
    def __init__(self, name, meta_data=None, custom_directory=None, cache_required=True,
                 create_directories=True, create_mode=0o755):
        ObjectWithMetaData.__init__(self, meta_data)
        self.custom_directory = custom_directory
        self.name = name
        self.create_directories = create_directories
        self.create_mode = create_mode
        self.cache_required = cache_required
    def __repr__(self): return self.repr
    @pimms.value
    def repr(name):
        '''
        dataset.repr is the representation string used for the given dataset.
        '''
        return ("Dataset('%s')" % name) if pimms.is_str(name) else ("Dataset%s" % (name,))
    @staticmethod
    def to_name(nm):
        '''
        Dataset.to_name(name) yields a valid dataset name equivalent to the given name or raises an
          error if name is not valid. In order to be valid, a name must be either strings or a tuple
          of number and strings that start with a string.
        '''
        if pimms.is_str(nm): return nm
        if not pimms.is_vector(nm): raise ValueError('name must be a string or tuple')
        if len(nm) < 1: raise ValueError('names that are tuples must have at least one element')
        if not pimms.is_str(nm): raise ValueError('names that are tuples must begin with a string')
        if not all(pimms.is_str(x) or pimms.is_number(x) for x in nm):
            raise ValueError('dataset names that are tuples must contain only strings and numbers')
        return tuple(nm)
    @pimms.param
    def custom_directory(d):
        '''
        dataset.custom_directory is None if no custom directory was provided for the given dataset;
          otherwise it is the provided custom directory.
        '''
        if d is None: return None
        if not pimms.is_str(d): raise ValueError('custom_directory must be a string')
        else: return d
    @pimms.param
    def create_directories(c):
        '''
        dataset.create_directories is True if the dataset was instructed to create its cache
        directory, should it be found to not exist, and is otherwise False.
        '''
        return bool(c)
    @pimms.param
    def create_mode(c):
        '''
        dataset.create_mode is the octal permision mode used to create the cache directory for the
        given dataset, if the dataset had to create its directory at all.
        '''
        return c
    @pimms.param
    def name(nm):
        '''
        dataset.name is either a string or a tuple of strings and numbers that identifies the given
        dataset. If dataset.name is a tuple, then the first element must be a string.
        '''
        return Dataset.to_name(nm)
    @pimms.param
    def cache_required(cr):
        '''
        dataset.cache_required is True if the dataset requires a cache directory and False
        otherwise.
        '''
        return cr
    @pimms.value
    def cache_root(custom_directory):
        '''
        dataset.cache_root is the root directory in which the given dataset has been cached.
        '''
        if custom_directory is not None: return None
        elif config['data_cache_root'] is None:
            # we create a data-cache in a temporary directory
            path = tempfile.mkdtemp(prefix='npythy_data_cache_')
            if not os.path.isdir(path): raise ValueError('Could not find or create cache directory')
            config['data_cache_root'] = path
            atexit.register(shutil.rmtree, path)
        return config['data_cache_root']
    @pimms.value
    def cache_directory(cache_root, name, custom_directory):
        '''
        dataset.cache_directory is the directory in which the given dataset is cached.
        '''
        if custom_directory is not None: return custom_directory
        return os.path.join(cache_root, (name    if pimms.is_str(name) else
                                         name[0] if len(name) == 1     else
                                         '%s_%x' % (name[0], hash(name[1:]))))
    @pimms.require
    def ensure_cache_directory(cache_directory, create_directories, create_mode, cache_required):
        '''
        ensure_cache_directory requires that a dataset's cache directory exists and raises an error
        if it cannot be found.
        '''
        if not cache_required: return True
        if os.path.isdir(cache_directory): return True
        if not create_directories:
            raise ValueError('dataset cache directory not found: %s' % (cache_directory,))
        os.makedirs(os.path.abspath(cache_directory), create_mode)
        return True
# We create the dataset repository: this is just a lazy map; to add a dataset to iit, use the
# function add_dataset(), immediately below
data = pimms.lazy_map({})
def add_dataset(dset, fn=None):
    '''
    add_dataset(dset) adds the given dataset to the neuropythy.data map.
    add_dataset(name, fn) adds a dataset with the given name; fn must be a function of zero
      arguments that yields the dataset.

    add_dataset always yeilds None or raises an error.
    '''
    global data
    if fn is None:
        if not isinstance(dset, Dataset):
            raise ValueError('Cannot add non-Dataset object to neuropythy datasets')
        nm = dset.name
        data = data.set(nm, dset)
    else:
        nm = Dataset.to_name(dset)
        def _load_dset():
            x = fn()
            if not isinstance(x, Dataset):
                raise ValueError('Loader for dataset %s failed to return a dataset' % nm)
            return x
        data = data.set(nm, _load_dset)
    # we want to update neuropythy.data also; this is a bit of a hack, but should work fine
    import neuropythy
    neuropythy.data = data
    return None


            
        



####################################################################################################
# Visual Performance Fields Dataset Description
# This file implements a neuropythy dataset object fo the visual performance
# fields project by Benson, Kupers, Barbot, Carrasco, and Winawer.
# By Noah C. Benson

import neuropythy, pimms

from ..util import (config)
from .core  import (Dataset, add_dataset)
from .hcp   import (to_nonempty_path, HCPMetaDataset)

config.declare('visual_performance_fields_path',
               environ_name='VISUAL_PERFORMANCE_FIELDS_PATH',
               default_value=None, filter=to_nonempty_path)
@pimms.immutable
class VisualPerformanceFieldsDataset(HCPMetaDataset):
    '''
    The VisualPerformanceFieldsDataset class declares the basic data structures and the functions
    for generating, loading, or preprocessing the visual performance fields dataset associated
    with the paper by Benson, Kupers, Carrasco, and Winawer (2021) in the journal eLife. The 
    dataset is typically accessed via the neuropythy library:
    ```
    import neuropythy as ny
    ny.data['visual_performance_fields']
    ```
    
    This dataset contains the following members:
      * subject_list is a list of the subject IDs for each HCP subject analyzed in this notebook.
      * roi_angles is a tuple of angle-widths for the distance-based ROIs around the V1/V2
        boundary or around the V1 horizontal meridian.
      * roi_eccens is a tuple of 2-tuples, each of which contains the (min_eccen, max_eccen) of
        one eccentricity ring used in defining the distance-based ROIs.
      * roi_angles_behavior and roi_eccens_behavior store the angles and eccentricities of the
        behavior-matched distance-based ROIs examined in the paper.
      * roi_angles_silva2018 and roi_eccens_silva2018 store the angles and eccentricities of the
        distance-based ROIs that are matched to the Silva et al. (2018) paper.
      * osf_url is the Open Science Framework URL for the project; this isn't an official URL, but
        it has the format "osf://XXXXX/" where the XXXXX is the OSF-id of the project.
      * metadata_table is a pandas dataframe of all the meta-data (i.e., the HCP behavioral data)
        for each subject.
      * gender is a dictionary whose keys are subject IDs and whose values are either 'M' or 'F'.
      * agegroup is a dictionary of age-groups for each subject; the age-group value is the age in
        the middleof the age range for a particular subject. E.g., if  subject u is in the 26-30
        age-group, then agegroup[u] will be 28.
      * inferred_maps is a dictionary mapping subject IDs to retinotopic mapping properties as
        inferred by Bayesian inference (Benson and Winawer, 2018, DOI:10.7554/eLife.40224). This
        dictionary is not used in the paper, but the data were used in analyses from earlier drafts
        of the manuscript, so they are included.
      * subjects is a dictionary of subject objects, each of which contain, as properties or
        meta-data, the sector information for each hemisphere.
      * DROI_table is a dataframe of information about the distance-based ROIs examined in the
        paper.
      * DROI_behavior_table and DROI_silva2018_table are equivalent distance-based ROI dataframes
        that are matched to the Barbot et al. (2020) data or the Silva et al. (2018) data.
      * asymmetry is a nested data-structure that contains informationn about the asymmetry of the
        ROIs examined in the paper.
      * asymmetry_table is a dataframe of information about the asymmetries for each subject.
      * barbot2020_data is a dictionary of data from Barbot et al. (2020) that is compared to the
        surface areas of the distance-based ROIs.
      * abrams2012_data is a dictionary of data from Abrams et al. (2012).
    '''
    
    # Constants / Magic Numbers ####################################################################
    roi_angles = (10, 20, 30, 40, 50)
    roi_angles_behavior = (7.5, 22.5, 37.5, 45)
    roi_angles_silva2018 = (5.625, 16.875,  28.125,  39.375, 45)
    roi_eccens = ((1,2), (2,3), (3,4), (4,5), (5,6), (6,7))
    roi_eccens_behavior = (4, 5)
    roi_eccens_silva2018 = ((1,2), (2,3), (3,4), (4,5), (5,6))

    # The eccentricity ranges we look at:
    # The URL for the OSF page from which we download our data.
    osf_url = 'osf://5gprz/'

    def __init__(self, angles=None, eccens=None, url=None, cache_directory=Ellipsis,
                 metadata_path=None, genetic_path=None, behavioral_path=None,
                 meta_data=None, create_directories=True, create_mode=0o755):
        cdir = cache_directory
        if cdir is Ellipsis:
            cdir = config['visual_performance_fields_path']
        HCPMetaDataset.__init__(self, name='visual_performance_fields',
                                cache_directory=cdir,
                                meta_data=meta_data,
                                create_mode=create_mode,
                                create_directories=create_directories,
                                cache_required=True,
                                metadata_path=metadata_path,
                                genetic_path=genetic_path,
                                behavioral_path=behavioral_path)
        self.url = url
        self.angles = angles
        self.eccens = eccens

    # Administrative Data ##########################################################################
    # Things that are required to load or organize the data but that aren't the data themselves.
    @pimms.param
    def url(url):
        '''
        url is the URL from which the performance-fields data is loaded.
        '''
        if url is None or url is Ellipsis: return VisualPerformanceFieldsDataset.osf_url
        if not pimms.is_str(u): raise ValueError('url must be a string')
        return u
    @pimms.value
    def pseudo_path(url, cache_directory):
        '''
        pseudo_path is the neuropythy psueod-path object responsible for loading the OSF data
        related to the visual performance fields dataset.
        '''
        from neuropythy.util import pseudo_path
        return pseudo_path(url, cache_path=cache_directory).persist()
    inferred_map_files = {'polar_angle':  '%s.%s.inf-MSMAll_angle.native32k.mgz',
                          'eccentricity': '%s.%s.inf-MSMAll_eccen.native32k.mgz',
                          'radius':       '%s.%s.inf-MSMAll_sigma.native32k.mgz',
                          'visual_area':  '%s.%s.inf-MSMAll_varea.native32k.mgz'}
    @pimms.value
    def hcp_data():
        '''
        hcp_data is the HCP Retinotopy dataset that is used in conjunction with the Visual
        Performance Fields dataset.
        '''
        return neuropythy.data['hcp_lines']
    @pimms.value
    def subject_list(hcp_data):
        '''
        subjcet_list is the list of subjects used in the dataset.
        '''
        import numpy as np
        # We want to filter out the subjects without mean sets of hand-drawn lines.
        excl = hcp_data.exclusions
        sids = [sid for sid in hcp_data.subject_list
                if ('mean', sid, 'lh') not in excl
                if ('mean', sid, 'rh') not in excl]
        sids = np.sort(sids)
        return tuple(sids)

    # Supporting Data ##############################################################################
    # Data that is critical for calculating the wedge ROIs but that aren't themselves the data used
    # to make the paper plots.
    @pimms.value
    def inferred_maps(pseudo_path, subject_list):
        '''
        inferred_maps is a nested-dictionary structure containing the retinotopic maps inferred by
        using Bayesian inference on the retinotopic maps of the subjects in the HCP 7T Retinotopy
        Dataset.
        '''
        import os, six
        from neuropythy.util import curry
        from neuropythy import load
        inffiles = VisualPerformanceFieldsDataset.inferred_map_files
        def _load_infmaps(sid,h,patt):
            flnm = pseudo_path.local_path('inferred_maps', patt % (sid,h))
            return load(flnm)
        return pimms.persist(
            {sid: {h: pimms.lmap({('inf_'+k): curry(_load_infmaps, sid, h, v)
                                  for (k,v) in six.iteritems(inffiles)})
                   for h in ['lh','rh']}
             for sid in subject_list})
    @pimms.param
    def angles(angs):
        '''
        angles is the set of polar angle deltas that are calculated. By default this is
          (10, 20, 30, 40, 50).
        '''
        if angs is None or angs is Ellipsis: return VisualPerformanceFieldsDataset.roi_angles
        else: return tuple(angs)
    @pimms.param
    def eccens(eccs):
        '''
        eccens is the set of polar angle deltas that are calculated. By default this is
          ((1,2), (2,3), (3,4), (4,5), (5,6), (6,7))
        '''
        import numpy as np
        if eccs is None or eccs is Ellipsis: eccs = VisualPerformanceFieldsDataset.roi_eccens
        sh = np.shape(eccs)
        if len(sh) == 2 and sh[1] == 2:
            eccs = sorted(eccs, key=lambda x:x[0])
            assert all(u[1] == v[0] for (u,v) in zip(eccs[:-1], eccs[1:])), \
                'eccens must be contiguous tuples'
            return tuple([tuple(u) for u in eccs])
        elif len(sh) == 1:
            eccs = sorted(eccs)
            return tuple(zip(eccs[:-1], eccs[1:]))
        else: raise ValueError("eccens must be a list of 2-tuples or a tuples of values")
    @pimms.value
    def sector_angles(angles):
        '''
        The angles used to calculate the sectors from which cortical magnifications are calculated.
        '''
        import numpy as np
        angles = np.asarray(angles)
        angles = np.concatenate([angles, 90 - angles, 90 + angles, 180 - angles, [0, 90, 180]])
        angles = np.unique(angles)
        angles.setflags(write=False)
        return angles
    @pimms.value
    def sector_eccens(eccens):
        '''
        The eccens used to calculate the sectors from which cortical magnifications are calculated.
        '''
        import numpy as np
        eccens = np.unique(eccens)
        eccens.setflags(write=False)
        return eccens
    @staticmethod
    def _generate_sector(hcp_data, sid, h, sector_angles, sector_eccens):
        import pyrsistent as pyr
        sects = hcp_data.refit_sectors(sid, h, sector_angles, sector_eccens)
        sects = {k+1: pyr.pmap(s) for (k,s) in enumerate(sects)}
        for uu in sects.values():
            for u in uu.values():
                u.setflags(write=False)
        return sects
    @pimms.value
    def sectors(hcp_data, sector_angles, sector_eccens, subject_list):
        '''
        sectors is a nested lazy-map whose first-level keys are subject ids and whose second-level
          keys are either 'lh' or 'rh'. The values are the dictionaries of sectors for each
          hemisphere, divided according to sector_angles and sector_eccens.
        '''
        import pyrsistent as pyr, neuropythy as ny
        f = VisualPerformanceFieldsDataset._generate_sector
        return pyr.pmap(
            {sid: pimms.lmap(
                {h: ny.util.curry(f, hcp_data, sid, h, sector_angles, sector_eccens)
                 for h in ('lh','rh')})
             for sid in subject_list})
    @pimms.value
    def sector_key(sectors, sector_angles, sector_eccens):
        '''
        sector_key is a 2-tuple of persistent maps; the first gives the sector label when queried
          by the sector boundary ids (i.e., 4-tuples of the min and max angles) and the second gives
          the inverse.
        '''
        import pyrsistent as pyr
        key = {}
        ii = 0
        for va in [1,2,3]:
            for (a0,a1) in zip(sector_angles[:-1], sector_angles[1:]):
                for (e0,e1) in zip(sector_eccens[:-1], sector_eccens[1:]):
                    ii += 1
                    key[(va,a0,a1,e0,e1)] = ii
        inv = {ii:k for (k,ii) in key.items()}
        return (pyr.pmap(key), pyr.pmap(inv))
    @staticmethod
    def _generate_sector_labels(hcp_data, sid, h, sectors, key):
        import numpy as np
        hem = hcp_data.subjects[sid].hemis[h]
        sct = sectors[sid][h]
        lbls = np.zeros(hem.vertex_count, dtype=np.int)
        for (tup,k) in key.items():
            if not pimms.is_tuple(tup):
                print(tup, k)
            (va,tup) = (tup[0], tup[1:])
            lbls[sct[va][tup]] = k
        lbls.setflags(write=False)
        return lbls
    def generate_sector_labels(self, sid, h, bins='default'):
        '''
        Generates and yields the sector labels for the given subject id and hemisphere; the optional
          third argument bins can be set to 'behavior' for the behaviorally-matched sector labels or
          to 'silva2018' for the ROIs used in Silva et al. (2018).
        '''
        f = VisualPerformanceFieldsDataset._generate_sector_labels
        (scts,skey) = (
            (self.behavior_sectors, self.behavior_sector_key)   if bins == 'behavior'  else
            (self.silva2018_sectors, self.silva2018_sector_key) if bins == 'silva2018' else
            (self.sectors, self.sector_key)                     if bins == 'default'   else
            (None, None))
        if skey is None: raise ValueError('Unrecognized bins parameter: %s' % (bins,))
        skey = skey[0]
        return f(self.hcp_data, sid, h, scts, skey)
    @staticmethod
    def _sector_labels(hcp_data, pseudo_path, tag, sid, h, sectors, key):
        import neuropythy as ny, numpy as np
        try:                cpath = pseudo_path.local_path('sectors%s' % tag, '%s.%s.mgz' % (h,sid))
        except ValueError:  cpath = None
        if cpath is None:
            return VisualPerformanceFieldsDataset._generate_sector_labels(
                hcp_data, sid, h, sectors, key)
        else:
            lbl = np.array(ny.load(cpath))
            lbl.setflags(write=False)
            return lbl
    @pimms.value
    def sector_labels(hcp_data, pseudo_path, sectors, sector_key):
        '''
        sector_labels is a nested lazy-map whose first-level keys are subject ids and whose
          second-level keys are hemisphere names 'lh' or 'rh'. The values are property maps
          for the associated subject's hemisphere's sector labels. The label key is given
          by sector_key.
        '''
        import pyrsistent as pyr, neuropythy as ny
        f = VisualPerformanceFieldsDataset._sector_labels
        skey = sector_key[0]
        return pyr.pmap(
            {sid: pimms.lmap(
                {h: ny.util.curry(f, hcp_data, pseudo_path, '', sid, h, sectors, skey)
                 for h in ('lh','rh')})
             for sid in sectors.keys()})
    @pimms.value
    def behavior_sectors(hcp_data, subject_list):
        '''
        behavior_sectors is a nested lazy-map like sectors whose first-level keys are subject ids
          and whose second-level keys are either 'lh' or 'rh'. The values are the dictionaries of
          sectors for each hemisphere, divided according to sector_angles and sector_eccens. Unlike
          sectors, behavior_sectors uses the behavior-matched ROIs.
        '''
        import numpy as np, pyrsistent as pyr, neuropythy as ny
        CLS = VisualPerformanceFieldsDataset
        angs = np.array(CLS.roi_angles_behavior)
        sector_angles = np.concatenate([angs, 90 - angs, 90 + angs, 180 - angs, [0, 90, 180]])
        sector_angles = np.unique(sector_angles)
        sector_eccens = np.unique(CLS.roi_eccens_behavior)
        f = lambda sid,h: CLS._generate_sector(hcp_data, sid, h, sector_angles, sector_eccens)
        return pyr.pmap({sid: pimms.lmap({h: ny.util.curry(f, sid, h) for h in ('lh','rh')})
                         for sid in subject_list})
    @pimms.value
    def behavior_sector_key(behavior_sectors):
        '''
        behavior_sector_key is a 2-tuple of persistent maps; the first gives the sector label when
          queried by the sector boundary ids (i.e., 4-tuples of the min and max angles) and the
          second gives the inverse.
        '''
        import pyrsistent as pyr, numpy as np
        CLS = VisualPerformanceFieldsDataset
        angs = np.asarray(CLS.roi_angles_behavior)
        sector_angles = np.concatenate([angs, 90 - angs, 90 + angs, 180 - angs, [0, 90, 180]])
        sector_angles = np.unique(sector_angles)
        sector_eccens = np.unique(CLS.roi_eccens_behavior)
        key = {}
        ii = 0
        for va in [1,2,3]:
            for (a0,a1) in zip(sector_angles[:-1], sector_angles[1:]):
                for (e0,e1) in zip(sector_eccens[:-1], sector_eccens[1:]):
                    ii += 1
                    key[(va,a0,a1,e0,e1)] = ii
        inv = {ii:k for (k,ii) in key.items()}
        return (pyr.pmap(key), pyr.pmap(inv))
    @pimms.value
    def behavior_sector_labels(hcp_data, pseudo_path, behavior_sectors, behavior_sector_key):
        '''
        behavior_sector_labels is a nested lazy-map like sector_labels whose first-level keys
          are subject ids and whose second-level keys are hemisphere names 'lh' or 'rh'. The
          values are property maps for the associated subject's hemisphere's sector labels. The
          label key is given by sector_key.
        '''
        import pyrsistent as pyr, neuropythy as ny
        f = VisualPerformanceFieldsDataset._sector_labels
        skey = behavior_sector_key[0]
        return pyr.pmap(
            {sid: pimms.lmap(
                {h: ny.util.curry(f, hcp_data, pseudo_path, '_behavior',
                                  sid, h, behavior_sectors, skey)
                 for h in ('lh','rh')})
             for sid in behavior_sectors.keys()})
    @staticmethod
    def _generate_hemi(sub, sid, h, infmaps, lbls, bhv_lbls):
        import neuropythy as ny, six
        hem = sub.hemis[h]
        bi = infmaps[sid][h]
        lbls = lbls[sid][h]
        bhv_lbls = bhv_lbls[sid][h]
        ps = {'vpf_sector':lbls, 'vpf_behavior_sector': bhv_lbls}
        for (k,v) in six.iteritems(bi): ps[k] = v
        return hem.with_prop(ps)
    @pimms.value
    def silva2018_sectors(hcp_data, subject_list):
        '''
        silva2018_sectors is a nested lazy-map like sectors whose first-level keys are subject ids
          and whose second-level keys are either 'lh' or 'rh'. The values are the dictionaries of
          sectors for each hemisphere, divided according to sector_angles and sector_eccens. Unlike
          sectors, silva2018_sectors uses the ROIs matched to Silva et al. (2018), Figure 4.
        '''
        import numpy as np, pyrsistent as pyr, neuropythy as ny
        CLS = VisualPerformanceFieldsDataset
        angs = np.array(CLS.roi_angles_silva2018)
        sector_angles = np.concatenate([angs, 90 - angs, 90 + angs, 180 - angs, [0, 90, 180]])
        sector_angles = np.unique(sector_angles)
        sector_eccens = np.unique(CLS.roi_eccens_silva2018)
        f = lambda sid,h: CLS._generate_sector(hcp_data, sid, h, sector_angles, sector_eccens)
        return pyr.pmap({sid: pimms.lmap({h: ny.util.curry(f, sid, h) for h in ('lh','rh')})
                         for sid in subject_list})
    @pimms.value
    def silva2018_sector_key(silva2018_sectors):
        '''
        silva2018_sector_key is a 2-tuple of persistent maps; the first gives the sector label when
          queried by the sector boundary ids (i.e., 4-tuples of the min and max angles) and the
          second gives the inverse.
        '''
        import pyrsistent as pyr, numpy as np
        CLS = VisualPerformanceFieldsDataset
        angs = np.asarray(CLS.roi_angles_silva2018)
        sector_angles = np.concatenate([angs, 90 - angs, 90 + angs, 180 - angs, [0, 90, 180]])
        sector_angles = np.unique(sector_angles)
        sector_eccens = np.unique(CLS.roi_eccens_silva2018)
        key = {}
        ii = 0
        for va in [1,2,3]:
            for (a0,a1) in zip(sector_angles[:-1], sector_angles[1:]):
                for (e0,e1) in zip(sector_eccens[:-1], sector_eccens[1:]):
                    ii += 1
                    key[(va,a0,a1,e0,e1)] = ii
        inv = {ii:k for (k,ii) in key.items()}
        return (pyr.pmap(key), pyr.pmap(inv))
    @pimms.value
    def silva2018_sector_labels(hcp_data, pseudo_path, silva2018_sectors, silva2018_sector_key):
        '''
        silva2018_sector_labels is a nested lazy-map like sector_labels whose first-level keys
          are subject ids and whose second-level keys are hemisphere names 'lh' or 'rh'. The
          values are property maps for the associated subject's hemisphere's sector labels. The
          label key is given by sector_key.
        '''
        import pyrsistent as pyr, neuropythy as ny
        f = VisualPerformanceFieldsDataset._sector_labels
        skey = silva2018_sector_key[0]
        return pyr.pmap(
            {sid: pimms.lmap(
                {h: ny.util.curry(f, hcp_data, pseudo_path, '_silva2018',
                                  sid, h, silva2018_sectors, skey)
                 for h in ('lh','rh')})
             for sid in silva2018_sectors.keys()})
    @staticmethod
    def _generate_subject(hcpdata, sid, infmaps, lbls, bhv_lbls):
        import neuropythy as ny, six
        sub = hcpdata.subjects[sid]
        f = VisualPerformanceFieldsDataset._generate_hemi
        lh = f(sub, sid, 'lh', infmaps, lbls, bhv_lbls)
        rh = f(sub, sid, 'rh', infmaps, lbls, bhv_lbls)
        return sub.with_hemi(lh=lh, rh=rh)
    @pimms.value
    def subjects(inferred_maps, sector_labels, behavior_sector_labels, subject_list, hcp_data):
        '''
        subjects is a dictionary of subject objects for all subjects used in the visual performance
        fields dataset. All subject objects in the subejcts dict include property data on the native
        hemispheres for inferred retinotopic maps and for V1 boundary distances.
        '''
        from neuropythy.util import curry
        f = VisualPerformanceFieldsDataset._generate_subject
        return pimms.lmap(
            {sid: curry(f, hcp_data, sid, inferred_maps, sector_labels, behavior_sector_labels)
             for sid in subject_list})

    # ROI-calculation Functions ####################################################################
    # These are methods that calculate the distance-based ROIs ("DROIs").
    @staticmethod
    def _generate_DROI(sector_labels, sector_key, ref_angle, angle_delta,
                       greater=True, lesser=True, visual_areas=(1,),
                       min_variance_explained=0, surface_area='midgray', eccentricity_range=(1,7)):
        import numpy as np
        ii = []
        skey = sector_key[0]
        if pimms.is_number(visual_areas): visual_areas = (visual_areas,)
        min_DROI_angle = ref_angle - angle_delta if lesser else ref_angle
        max_DROI_angle = ref_angle + angle_delta if greater else ref_angle
        min_DROI_eccen = eccentricity_range[0]
        max_DROI_eccen = eccentricity_range[1]
        for (tup,lbl) in skey.items():
            (va, mnang, mxang, mnecc, mxecc) = tup
            if va not in visual_areas: continue
            if mnang < min_DROI_angle: continue
            if mxang > max_DROI_angle: continue
            if mnecc < min_DROI_eccen: continue
            if mxecc > max_DROI_eccen: continue
            ii.append(np.where(sector_labels == lbl)[0])
        if len(ii) == 0:
            return []
        else:
            ii = np.concatenate(ii)
            return np.unique(ii)
    @staticmethod
    def generate_DROI(self, sid, h, ref_angle, angle_delta, bins='default',
                      greater=True, lesser=True, visual_areas=(1,),
                      min_variance_explained=0, surface_area='midgray', eccentricity_range=(1,7)):
        '''
        Generate a set of distance-based ROIs for the given subject and hemisphere from the given
          set of sector labels and sector key
        '''
        (sct,skey) = (
            (self.behavior_sector_labels,  self.behavior_sector_key)  if bins == 'behavior'  else
            (self.silva2018_sector_labels, self.silva2018_sector_key) if bins == 'silva2018' else
            (self.sector_labels,           self.sector_key)           if bins == 'default'   else
            (None,                         None))
        if skey is None: raise ValueError('Unrecognized bins parameter: %s' % (bins,))
        sct = sct[sid][h]
        return VisualPerformanceFieldsDataset._generate_DROI(
            sub, h, sct, skey,
            greater=greater, lesser=lesser, visual_areas=visual_areas,
            min_variance_explained=min_variance_explained,
            surface_area=surface_area,
            eccentricity_range=eccentricity_range)
    @staticmethod
    def _vertical_DROI_from_ventral_dorsal(vnt, drs):
        '''
        Given vnt and drs return values from the generate_DROI_data() function,
        yields the 'vertical' ROI (the V1 parts of vnt and drs) of the combined ROI.
        '''
        import neuropythy as ny, numpy as np, six
        (iiv, iid) = [np.where(q['visual_area'] == 1)[0] for q in (vnt,drs)]
        if len(iiv) == 0 and len(iid) == 0: return {k:[] for k in six.iterkeys(drs)}
        if len(iiv) == 0: return {k:v[iid] for (k,v) in six.iteritems(drs)}
        if len(iid) == 0: return {k:v[iiv] for (k,v) in six.iteritems(vnt)}
        res = {}
        for (k,v) in six.iteritems(vnt):
            res[k] = np.concatenate([v[iiv], drs[k][iid]])
        return res
    @staticmethod
    def _generate_subject_DROI_boundary_data(sub, h, paradigm, angle_delta,
                                             sector_label, sector_key,
                                             min_variance_explained=0,
                                             surface_area='midgray',
                                             eccentricity_range=(1,7)):
        import neuropythy as ny, numpy as np, copy, six
        CLS = VisualPerformanceFieldsDataset
        paradigm = paradigm.lower()
        if paradigm == 'vertical':
            # This is the weird case: we handle it separately: just run the function
            # for both ventral and dorsal and concatenate the V1 parts
            (vnt,drs) = [CLS._generate_subject_DROI_boundary_data(sub, h, DROIs, para, angle_delta)
                         for para in ['ventral', 'dorsal']]
            return CLS._vertical_DROI_from_ventral_dorsal(vnt, drs)
        # Get the hemisphere.
        hem = sub.hemis[h]
        # Some things depend on the paradigm.
        if paradigm == 'ventral':
            center = 0
            gt = True
            lt = True
            vas = (1,2)
        elif paradigm == 'dorsal':
            center = 180
            gt = True
            lt = True
            vas = (1,2)
        elif paradigm == 'horizontal':
            center = 90
            gt = True
            lt = True
            vas = (1,)
        elif paradigm == 'hventral':
            center = 90
            gt = False
            lt = True
            vas = (1,)
        elif paradigm == 'hdorsal':
            center = 90
            gt = True
            lt = False
            vas = (1,)
        elif paradigm == 'ventral_v1':
            center = 0
            gt = True
            lt = False
            vas = (1,)
        elif paradigm == 'dorsal_v1':
            center = 180
            gt = False
            lt = True
            vas = (1,)
        elif paradigm == 'ventral_v2':
            center = 0
            gt = True
            lt = False
            vas = (2,)
        elif paradigm == 'dorsal_v2':
            center = 180
            gt = False
            lt = True
            vas = (2,)
        else: raise ValueError('unrecognized paradigm: %s' % (paradigm,))
        # Get the indices
        f = VisualPerformanceFieldsDataset._generate_DROI
        ii = f(sector_label, sector_key,
               center, angle_delta,
               greater=gt, lesser=lt,
               visual_areas=vas,
               min_variance_explained=min_variance_explained,
               surface_area=surface_area,
               eccentricity_range=eccentricity_range)
        # Grab the other data:
        surface_area = hem.property('midgray_surface_area')
        white_surface_area = hem.property('white_surface_area')
        pial_surface_area = hem.property('pial_surface_area')
        sa = surface_area[ii]
        wsa = white_surface_area[ii]
        psa = pial_surface_area[ii]
        th = hem.prop('thickness')[ii]
        vl = sa * th
        return {'surface_area_mm2': sa,
                'white_surface_area_mm2': wsa,
                'pial_surface_area_mm2': psa,
                'mean_thickness_mm':th,
                'volume_mm3': vl,
                'indices': ii,
                'visual_area': hem.prop('visual_area')[ii]}
    def generate_subject_DROI_boundary_data(self, sid, h, paradigm, angle_delta,
                                            bins='default',
                                            min_variance_explained=0,
                                            surface_area='midgray',
                                            eccentricity_range=(1,7)):
        '''
        generate_subject_DROI_boundary_data(sid, h, DROIs, paradigm, delta) yields a dict of data
          about the given distance ROI; the ROI is defined by the paradigm, which must be one of
          'vertical', 'horizontal', 'ventral', or 'dorsal', and delta, which is the distance in
          polar angle degrees from the given boundary.
        '''
        f = VisualPerformanceFieldsDataset._generate_subject_DROI_boundary_data
        (sct,skey) = (
            (self.behavior_sector_labels,  self.behavior_sector_key)  if bins == 'behavior'  else
            (self.silva2018_sector_labels, self.silva2018_sector_key) if bins == 'silva2018' else
            (self.sector_labels,           self.sector_key)           if bins == 'default'   else
            (None, None))
        if skey is None: raise ValueError('Unrecognized bins parameter: %s' % (bins,))
        sct = sct[sid][h]
        return f(self.subjects[sid], h, sct, skey, paradigm, angle_delta,
                 min_variance_explained=min_variance_explained,
                 surface_area=surface_area,
                 eccentricity_range=eccentricity_range)
    @staticmethod
    def _generate_subject_DROI_data(sub, h, sector_labels, sector_key, angle_delta,
                                    results='summary',
                                    min_variance_explained=0,
                                    surface_area='midgray',
                                    eccentricity_range=(1,7)):
        import neuropythy as ny, numpy as np, six
        paradigms = ['ventral','dorsal','horizontal','hdorsal','hventral',
                     'ventral_v1','ventral_v2','dorsal_v1','dorsal_v2']
        results = results.lower()
        kw = dict(min_variance_explained=min_variance_explained,
                  surface_area=surface_area,
                  eccentricity_range=eccentricity_range)
        # Reorganize these into the paradigm-based ROIs.
        f = VisualPerformanceFieldsDataset._generate_subject_DROI_boundary_data
        (vnt,drs,hrz,hdrs,hvnt,vnt1,vnt2,drs1,drs2) = [
            f(sub, h, para, angle_delta, sector_labels, sector_key, **kw)
            for para in paradigms]
        # we don't need to run vertical because we can derive it from the other measures:
        ver = VisualPerformanceFieldsDataset._vertical_DROI_from_ventral_dorsal(vnt, drs)
        # depending on the results arg, we return these or their summaries
        res = {'vertical': ver, 'horizontal': hrz, 'ventral': vnt, 'dorsal': drs,
               'hdorsal': hdrs, 'hventral': hvnt, 'ventral_v1': vnt1, 'dorsal_v1': drs1,
               'ventral_v2': vnt2, 'dorsal_v2': drs2}
        if results == 'summary':
            fix = {'surface_area_mm2': np.nansum,
                   'white_surface_area_mm2': np.nansum,
                   'pial_surface_area_mm2': np.nansum,
                   'volume_mm3': np.nansum,
                   'mean_thickness_mm': lambda x: np.nan if len(x) == 0 else np.nanmean(x)}
            return {k: {kk: fix[kk](vv) for (kk,vv) in six.iteritems(v) if kk in fix}
                    for (k,v) in six.iteritems(res)}
        else:
            return res
    def generate_subject_DROI_data(self, sid, h, angle_delta, results='summary',
                                   bins='default',
                                   min_variance_explained=0,
                                   surface_area='midgray',
                                   eccentricity_range=(1,7)):
        '''
        generate_subject_DROI_data(sid, h, angle_delta) yields distance-based ROI data for
          the set of ROIs (ventral, dorsal, horizontal, vertical) for the given
          subject and hemisphere.
        '''
        (sct,skey) = (
            (self.behavior_sector_labels,  self.behavior_sector_key)  if bins == 'behavior'  else
            (self.silva2018_sector_labels, self.silva2018_sector_key) if bins == 'silva2018' else
            (self.sector_labels,           self.sector_key)           if bins == 'default'   else
            (None, None))
        if skey is None: raise ValueError('Unrecognized bins parameter: %s' % (bins,))
        sct = sct[sid][h]
        return VisualPerformanceFieldsDataset._generate_subject_DROI_data(
            self.subjects[sid], h, sct, skey, angle_delta, results=results,
            min_variance_explained=min_variance_explained,
            surface_area=surface_area,
            eccentricity_range=eccentricity_range)
    @staticmethod
    def _generate_subject_DROI_table(subjects, sid, sector_labels, sector_key,
                                     min_variance_explained=0,
                                     surface_area='midgray'):
        import neuropythy as ny, numpy as np, six
        angles = np.unique([k for kk in sector_key[0].keys() for k in kk[1:3]])
        angles = angles[:next(ii+1 for (ii,ang) in enumerate(angles) if ang >= 45)]
        if angles[0] == 0: angles = angles[1:]
        eccens = np.unique([k for kk in sector_key[0].keys() for k in kk[3:]])
        eccens = list(zip(eccens[:-1], eccens[1:]))
        # For each eccentricity range, we'll start by making the DROIs, which will
        # use some sharerd parameters.
        kw = dict(min_variance_explained=min_variance_explained,
                  surface_area=surface_area)
        # We'll buiild up a table.
        tbl = ny.auto_dict(None, [])
        sub = subjects[sid]
        sct = sector_labels[sid]
        CLS = VisualPerformanceFieldsDataset
        for h in ['lh','rh']:
            # go through the eccen ranges and angles:
            for erng in eccens:
                kw['eccentricity_range'] = tuple(erng)
                for ang in angles:
                    # get all the summary data:
                    alldat = CLS._generate_subject_DROI_data(sub, h, sct[h], sector_key, ang, **kw)
                    for (para,dat) in six.iteritems(alldat):
                        # append to the appropriate columns:
                        tbl['sid'].append(sid)
                        tbl['hemisphere'].append(h)
                        tbl['boundary'].append(para)
                        tbl['min_eccentricity_deg'].append(erng[0])
                        tbl['max_eccentricity_deg'].append(erng[1])
                        tbl['angle_delta_deg'].append(ang)
                        for (k,v) in six.iteritems(dat):
                            tbl[k].append(v)
        return ny.to_dataframe(tbl)
    def generate_subject_DROI_table(self, sid, bins='default', min_variance_explained=0,
                                    surface_area='midgray'):
        """
        Calculate the distance-based ROIs for a single subject; indended for use with
        multiprocessing. This function will load the subject's data instead of running
        the calculation if the relevant data-file exists.
        """
        (sctlbl,skey) = (
            (self.behavior_sector_labels,  self.behavior_sector_key)  if bins == 'behavior'  else
            (self.silva2018_sector_labels, self.silva2018_sector_key) if bins == 'silva2018' else
            (self.sector_labels,           self.sector_key)           if bins == 'default'   else
            (None, None))
        if skey is None: raise ValueError('Unrecognized bins parameter: %s' % (bins,))
        f = VisualPerformanceFieldsDataset._generate_subject_DROI_table
        return f(self.subjects, sid, sctlbl, skey,
                 min_variance_explained=min_variance_explained,
                 surface_area=surface_area)
    @staticmethod
    def _generate_subject_DROI_details(subjects, sid, h, sector_labels, sector_key,
                                       eccentricity_range=None,
                                       min_variance_explained=0):
        from neuropythy.util import curry
        import six, pyrsistent as pyr, numpy as np
        paradigm_order = ['dorsal', 'ventral', 'vertical', 'horizontal']
        angles = np.unique([k for kk in sector_key[0].keys() for k in kk[1:3]])
        angles = angles[:next(ii+1 for (ii,ang) in enumerate(angles) if ang >= 45)]
        if eccentricity_range is None:
            e = np.unique([k for kk in sector_key[0].keys() for k in kk[3:]])
            e = list(zip(e[:-1], e[1:]))
        else:
            e = eccentricity_range
        if e is None or e is Ellipsis: e = eccens
        kw = dict(min_variance_explained=min_variance_explained)
        if (pimms.is_list(e) or pimms.is_tuple(e)) and all(pimms.is_tuple(q) for q in e):
            f = VisualPerformanceFieldsDataset._generate_subject_DROI_details
            res = [f(subjects, sid, h, sector_labels, sector_key, eccentricity_range=q, **kw)
                   for q in e]
            q = res[0]
            def _merge(p, a):
                r = {}
                for k in six.iterkeys(q[p][a]):
                    u = [u[p][a][k] for u in res if len(u[p][a][k]) > 0]
                    if len(u) == 0: u = np.asarray([], dtype=np.float)
                    else: u = np.concatenate(u)
                    r[k] = u
                return pyr.pmap(r)
            return pyr.pmap({k: pimms.lmap({a: curry(_merge, k, a) for a in angles})
                             for k in paradigm_order})
        sct = sector_labels[sid][h]
        sub = subjects[sid]
        kw['eccentricity_range'] = e
        kw['results'] = 'all'
        f = VisualPerformanceFieldsDataset._generate_subject_DROI_data
        gendeets = lambda sid, h, k: f(sub, h, sct, sector_key, k, **kw)
        lm0 = pimms.lmap({k: curry(gendeets, sid, h, k) for k in angles})
        pfn = lambda p: pimms.lmap({k:curry(lambda k:lm0[k][p], k) for k in angles})
        return pimms.lmap({p: curry(pfn, p) for p in paradigm_order})
    def generate_subject_DROI_details(self, sid, h, eccentricity_range=None, bins='default',
                                      min_variance_explained=0):
        """
        Calculate the distance-based ROI details for a single subject. Details contain
        similar data as the subject's DROI table, but is in a nested dictionary format.
        """
        f = VisualPerformanceFieldsDataset._generate_subject_DROI_details
        (sctlbl,skey) = (
            (self.behavior_sector_labels,  self.behavior_sector_key)  if bins == 'behavior'  else
            (self.silva2018_sector_labels, self.silva2018_sector_key) if bins == 'silva2018' else
            (self.sector_labels,           self.sector_key)           if bins == 'default'   else
            (None, None))
        if skey is None: raise ValueError('Unrecognized bins parameter: %s' % (bins,))
        return f(self.subjects, sid, h, sctlbl, skey,
                 eccentricity_range=eccentricity_range,
                 min_variance_explained=min_variance_explained)
    def generate_DROI_details(self, eccentricity_range=None, bins='default',
                              min_variance_explained=0):
        '''
        generate_DROI_details() yields a set of lazily computed DROI detailed analyses; these
        analyses are used to generate the DROI table(s).
        '''
        import six
        from neuropythy.util import curry
        f = VisualPerformanceFieldsDataset._generate_subject_DROI_details
        (sctlbl,skey) = (
            (self.behavior_sector_labels,  self.behavior_sector_key)  if bins == 'behavior'  else
            (self.silva2018_sector_labels, self.silva2018_sector_key) if bins == 'silva2018' else
            (self.sector_labels,           self.sector_key)           if bins == 'default'   else
            (None, None))
        if skey is None: raise ValueError('Unrecognized bins parameter: %s' % (bins,))
        m = {
            sid: pimms.lmap(
                {h: curry(f, self.subjects, sid, h, sctlbl, skey,
                          eccentricity_range=eccentricity_range,
                          min_variance_explained=min_variance_explained)
                 for h in ['lh','rh']})
            for sid in six.iterkeys(self.subjects)}
        return pimms.persist(m)
    @staticmethod
    def _generate_DROI_tables_call(tup):
        import os, neuropythy as ny, numpy as np
        # set our niceness!
        os.nice(10)
        # turn off numpy warnings.
        np.seterr(all='ignore')
        flnm = os.path.join(tup[3], '%s.csv' % (tup[0],))
        if os.path.isfile(flnm): return flnm
        f = VisualPerformanceFieldsDataset._generate_subject_DROI_table
        dd = ny.data['visual_performance_fields']
        subjects = dd.subjects
        bins = tup[1]
        (sctlbl,skey) = (
            (dd.behavior_sector_labels,  dd.behavior_sector_key)  if bins == 'behavior'  else
            (dd.silva2018_sector_labels, dd.silva2018_sector_key) if bins == 'silva2018' else
            (dd.sector_labels,           dd.sector_key)           if bins == 'default'   else
            (None, None))
        if skey is None: raise ValueError('Unrecognized bins parameter: %s' % (bins,))
        tbl = f(subjects, tup[0], sctlbl, skey, min_variance_explained=tup[2])
        ny.save(flnm, tbl)
        return flnm
    def generate_DROI_tables(self, bins='default', nprocs=None, printstatus=False,
                             min_variance_explained=0, method=None, tempdir=None):
        '''
        generate_DROI_tables() recalculates the set of distance-based ROIs for each subject
        based on the data in the inferred maps and pRFs.
        '''
        import neuropythy as ny, os, six, pyrsistent as pyr, multiprocessing as mp, numpy as np
        drois = {}
        subject_list = self.subject_list
        nsubs = len(subject_list)
        f = VisualPerformanceFieldsDataset._generate_DROI_tables_call
        if tempdir is None: tdir = ny.util.tmpdir()
        else: tdir = tempdir
        if nprocs is None or nprocs is Ellipsis: nprocs = mp.cpu_count()
        if nprocs < 2: nprocs = 1
        if nprocs > 1:
            for ii in np.arange(0, nsubs, nprocs):
                mx = np.min([len(subject_list), ii + nprocs])
                if printstatus:
                    print("%2d - %3d%%" % ((ii + 1)/nsubs * 100, mx/nsubs * 100))
                with mp.Pool(nprocs) as pool:
                    sids = subject_list[ii:ii+nprocs]
                    tbls = pool.map(f, [(sid, bins, min_variance_explained, tdir) for sid in sids])
                # add these data into the overall roi table
                for (sid,tbl) in zip(sids,tbls):
                    drois[sid] = ny.load(tbl)
        else:
            for (ii,sid) in enumerate(subject_list):
                if printstatus and ii % 10 == 9:
                    print('%3d (%d), %4.1f%%' % (ii, sid, ii/len(subject_list)*100.0))
                drois[sid] = f((sid, bins, min_variance_explained, tdir))
        return pyr.pmap(drois)
    def generate_DROI_table(self):
        '''
        generate_DROI_table() is a method that recalculates the DROI summary table from the
        individual subject DROIs.
        '''
        import pandas
        df = pandas.concat(list(self.DROI_tables.values()))
        df.set_index(['sid','hemisphere'])
        return df
    
    # Distance-based ROIs ##########################################################################
    # The distance-based/wedge ROIs themselves.
    @pimms.value
    def DROI_details(subjects, sector_labels, sector_key):
        '''
        DROI_details is a nested-dictionary structure of the various DROI details of each subject
        and hemisphere.
        '''
        import neuropythy as ny, os, six
        from neuropythy.util import curry
        f = curry(VisualPerformanceFieldsDataset._generate_subject_DROI_details, subjects)
        m = {sid: pimms.lmap({h: curry(f, sid, h, sector_labels, sector_key) for h in ['lh','rh']})
             for sid in six.iterkeys(subjects)}
        return pimms.persist(m)
    @pimms.value
    def DROI_tables(subject_list, pseudo_path):
        '''
        DROI_tables (distance-based regions of interest) is a dictionary of ROIS used in the visual
        performance field project. DROI_tables[sid] is a dataframe of the ROI-data for the subject
        with ID sid.
        '''
        import neuropythy as ny, os, six
        # Load one subject.
        def _load_DROI(sid):
            # get a subject-specific cache_path
            cpath = pseudo_path.local_path('DROIs', '%s.csv' % (sid,))
            return ny.load(cpath)
        return pimms.lmap({sid: ny.util.curry(_load_DROI, sid) for sid in subject_list})
    @pimms.value
    def DROI_table(pseudo_path):
        '''
        DROI_table (distance-based ROI table) is a dataframe summarizing all the data from all the
        hemispheres and all the distance-based wedge ROIs used in the visual performance fields
        project.
        '''
        import neuropythy as ny
        df = ny.load(pseudo_path.local_path('DROI_table.csv'))
        df.set_index(['sid','hemisphere'])
        return df
    @staticmethod
    def generate_DROI_summary(DROI_table, bins='default', exclude_periphery=True):
        '''
        generate_DROI_summary(table) converts the DROI table into a summary.
        '''
        import neuropythy as ny, numpy as np
        if bins == 'behavior':
            angles = VisualPerformanceFieldsDataset.roi_angles_behavior
            eccens = VisualPerformanceFieldsDataset.roi_eccens_behavior
        elif bins == 'silva2018':
            angles = VisualPerformanceFieldsDataset.roi_angles_silva2018
            eccens = VisualPerformanceFieldsDataset.roi_eccens_silva2018
        elif bins == 'default':
            angles = VisualPerformanceFieldsDataset.roi_angles
            eccens = VisualPerformanceFieldsDataset.roi_eccens
        else:
            raise ValueError('Unrecognized bins argument: %s' % (bins,))
        eccens = list(zip(eccens[:-1], eccens[1:]))
        # in eccens, by default, we exclude the peripheral (6-7 degree) eccentricity band.
        if exclude_periphery:
            eccens = [uv for uv in eccens if uv != (6,7)]
        emns = [ee[0] for ee in eccens]
        emxs = [ee[1] for ee in eccens]
        def _dfsel(df, k, ang, emns=[1,2,3,4,5], emxs=[2,3,4,5,6]):
            tbls = [ny.util.dataframe_select(df, angle_delta_deg=ang,
                                             min_eccentricity_deg=mn,
                                             max_eccentricity_deg=mx)
                    for (mn,mx) in zip(emns,emxs)]
            tbls = [tbl[['sid','hemisphere',k]] for tbl in tbls]
            tbl = tbls[0]
            for t in tbls[1:]:
                tt = tbl.merge(t, on=['sid','hemisphere'])
                tt[k] = tt[k+'_x'] + tt[k+'_y']
                tbl = tt[['sid','hemisphere',k]]
            tl = tbl.loc[tbl['hemisphere'] == 'lh']
            tr = tbl.loc[tbl['hemisphere'] == 'rh']
            tt = tl.merge(tr, on='sid')
            tt = tt.sort_values('sid')
            return tt[k+'_x'].values + tt[k+'_y'].values
        dat = {
            para: {
                k: pimms.imm_array([_dfsel(df, k, ang, emns=emns, emxs=emxs) for ang in angles])
                for k in ['surface_area_mm2', 'pial_surface_area_mm2', 'white_surface_area_mm2',
                          'mean_thickness_mm', 'volume_mm3']}
            for para in ['horizontal','vertical','dorsal','ventral','hdorsal','hventral',
                         'dorsal_v1','ventral_v1','dorsal_v2','ventral_v2']
            for df in [ny.util.dataframe_select(DROI_table, boundary=para)]}
        return pimms.persist(dat)
    @pimms.value
    def DROI_summary(DROI_table):
        '''
        DROI_summary (distance-based ROI summary) is a nested-dictionary data structure that
        provides easily-plottable summaries of the DROI_table. A value DROI_summary[b][k][u][s]
        corresponds to the boundary b ('ventral', 'dorsal', 'vertical', or 'horizontal'), the
        key k ('surface_area_mm2', 'mean_thickness_mm', or 'volume_mm3'), angle bin u (where 
        u = 0, 1, 2, 3, 4 indicates 0-10, 10-20, 20-30, 30-40, 40-50 degrees away from the relevant
        boundary), and subject s. Note that subject number is not necessarily consistent across
        boundaries and keys as some subjects have no vertices in small ROIs and thus get excluded
        from the relevant collections. All data are collapsed across eccentricities from 1-6
        degrees.
        '''
        return VisualPerformanceFieldsDataset.generate_DROI_summary(DROI_table)
    @pimms.value
    def DROI_behavior_tables(subject_list, pseudo_path):
        '''
        DROI_behavior_tables (distance-based regions of interest) is a dictionary of ROIS used in
        the visual performance field project. It is similar to DROI_tables, but it uses ROIs that
        are matched to the Barbot et al. (2020) measurements.
        '''
        import neuropythy as ny, os, six
        # Load one subject.
        def _load_DROI(sid):
            # get a subject-specific cache_path
            cpath = pseudo_path.local_path('DROIs_behavior', '%s.csv' % (sid,))
            return ny.load(cpath)
        return pimms.lmap({sid: ny.util.curry(_load_DROI, sid) for sid in subject_list})
    @pimms.value
    def DROI_behavior_table(pseudo_path):
        '''
        DROI_behavior_table (distance-based ROI table) is a dataframe summarizing all the data from
        all the hemispheres and all the distance-based wedge ROIs used in the visual performance fields
        project. Like DROI_table, but using the bejavior-matched DROIs.
        '''
        import neuropythy as ny
        df = ny.load(pseudo_path.local_path('DROI_table_behavior.csv'))
        df.set_index(['sid','hemisphere'])
        return df
    @pimms.value
    def DROI_behavior_summary(DROI_behavior_table):
        '''
        DROI_behavior_summary (distance-based ROI summary) is a nested-dictionary data structure
        like DROI_summary, but that uses the DROIs matched to the behavioral data from Barbot et al.
        (2020).
        '''
        return VisualPerformanceFieldsDataset.generate_DROI_summary(DROI_behavior_table,
                                                                    bins='behavior')
    @pimms.value
    def DROI_silva2018_tables(subject_list, pseudo_path):
        '''
        DROI_behavior_tables (distance-based regions of interest) is a dictionary of ROIS used in
        the visual performance field project. It is similar to DROI_tables, but it uses ROIs that
        are matched to Figure 4 from Silva et al. (2018).
        '''
        import neuropythy as ny, os, six
        # Load one subject.
        def _load_DROI(sid):
            # get a subject-specific cache_path
            cpath = pseudo_path.local_path('DROIs_silva2018', '%s.csv' % (sid,))
            return ny.load(cpath)
        return pimms.lmap({sid: ny.util.curry(_load_DROI, sid) for sid in subject_list})
    @pimms.value
    def DROI_silva2018_table(pseudo_path):
        '''
        DROI_silva2018_table (distance-based ROI table) is a dataframe summarizing all the data from
        all the hemispheres and all the distance-based wedge ROIs used in the visual performance
        fields project. Like DROI_table, but using the DROIs matched to Figure 4 of Silva et al.
        (2018).
        '''
        import neuropythy as ny
        df = ny.load(pseudo_path.local_path('DROI_table_silva2018.csv'))
        df.set_index(['sid','hemisphere'])
        return df
    @pimms.value
    def DROI_silva2018_summary(DROI_silva2018_table):
        '''
        DROI_silva2018_summary (distance-based ROI summary) is a nested-dictionary data structure
        like DROI_summary, but that uses the DROIs matched to the Figure 4 from Silva et al.
        (2018).
        '''
        return VisualPerformanceFieldsDataset.generate_DROI_summary(DROI_silva2018_table,
                                                                    bins='silva2018')
    @pimms.value
    def asymmetry(DROI_summary):
        '''
        asymmetry is a nested dictionary structure containing the surface-area asymmetry estimates
        for each subject. The value asymmetry[k][a][sno] is the percent asymmetry between the axes
        defined by comparison name k ('HMA' for HM:VM asymmetry, 'VMA' for LVM:UVM asymmetry,
        'HVA_cumulative' for cumulative HM:VM asymmetry, or 'VMA_cumulative' for cumulative LVM:UVM
        asymmetry), subject number sno (0-180 for the HCP subject whose ID is subject_list[sno]),
        and angle-distance a (10, 20, 30, 40, or 50 indicating the angle-wedge size in degrees of
        polar angle).

        Asymmetry is defined as (value1 - value2) / mean(value1, value2) where value1 and value2 are
        either the horizontal and vertical ROI surface areas respectively or the lower-vetical
        (dorsal) and upper-vertical (ventral) ROI surface areas respectively. The values reported
        in this data structure are percent asymmetry: difference / mean * 100.
        '''
        import neuropythy as ny, six, numpy as np
        quant = 'surface_area_mm2'
        dat = {}
        for (k,(k1,k2)) in zip(['HVA','VMA'], [('horizontal','vertical'), ('dorsal','ventral')]):
            for iscum in [True,False]:
                # Grab and prep the data.
                ys1 = np.asarray(DROI_summary[k1][quant])
                ys2 = np.asarray(DROI_summary[k2][quant])
                if not iscum:
                    res = []
                    for ys in (ys1,ys2):
                        (cum,yr) = (0, [])
                        for yy in ys:
                            yr.append(yy - cum)
                            cum = yy
                        res.append(yr)
                    (ys1,ys2) = [np.asarray(u) for u in res]
                # Calculate the asymmetries.
                asym = []
                for (y1,y2) in zip(ys1, ys2):
                    mu = np.nanmean([y1, y2], axis=0)
                    dy = y1 - y2
                    asym.append(dy * ny.math.zinv(mu) * 100)
                # Append the data
                dat[k + '_cumulative' if iscum else k] = pimms.imm_array(asym)
        return pimms.persist(dat)
    @pimms.value
    def asymmetry_table(asymmetry, agegroup, gender, subject_list):
        '''
        asymmetry_table is a pandas dataframe of asymmetry data for each subject
        that is organized by age-group and gender. In order for this table to
        be available, you must have access to the HCP restricted data, and 
        neuropythy must be configured to use it; otherwise an error will be
        raised.
        '''
        import neuropythy as ny
        import numpy as np
        # We can start by grabbing the subject ages and genders in the same order as
        # is used in data.asymmetry, which tracks the HVA and VMA of each subject.
        # This order is the same as that given in data.subject_list.
        agegroups = [agegroup[sid] for sid in subject_list]
        genders   = [gender[sid]   for sid in subject_list]

        # Now we just make a dataframe; all the columns are already ordered by subject
        # ID so we can just put them together. We do, however, need to expand each
        # column by the wedge-size because we are flattening that dimension out into
        # a single column.
        angle_col = 'ROI Width [deg polar angle]'
        age_col = 'Age-Group [years]'
        gender_col = 'Gender'
        data_cols = ['Local HVA', 'Cumulative HVA', 'Local VMA', 'Cumulative VMA']
        data_cols = [d + ' [%]' for d in data_cols]
        data_keys = ['HVA', 'HVA_cumulative', 'VMA', 'VMA_cumulative']
        nsubs = len(subject_list)
        
        df = {}
        roi_angles = VisualPerformanceFieldsDataset.roi_angles
        df[angle_col]  = np.reshape([np.full(nsubs, k) for k in roi_angles], -1)
        df['sid']      = np.reshape([subject_list for k in roi_angles], -1)
        df[gender_col] = np.reshape([genders           for k in roi_angles], -1)
        df[age_col]    = np.reshape([agegroups         for k in roi_angles], -1)
        for (col,k) in zip(data_cols, data_keys):
            df[col] = asymmetry[k].flatten()
        df = ny.to_dataframe(df)
        df = df[['sid',age_col,gender_col,angle_col] + data_cols]
        df.set_index('sid')
        return df
    @pimms.value
    def barbot2020_data(pseudo_path):
        '''
        Load and return the data from Barbot et al. (2020).
        '''
        import os, numpy as np, pyrsistent as pyr
        from scipy.io import loadmat
        filename = pseudo_path.local_path('supp', 'barbot_xue_carrasco_2020.mat')
        pp_raw = loadmat(filename)
        pp_dat = pp_raw['DATA']
        pp_dat = {nm: pp_dat[nm][0,0].T for nm in pp_dat.dtype.names}
        behavior_mtx = 10**pp_dat['SF_75THRESH_ALL']
        behavior_mean = np.mean(behavior_mtx, axis=1)
        behavior_sem = np.sqrt(np.var(behavior_mtx, axis=1) / behavior_mtx.shape[1])
        behavior_angs = pp_dat['ANGLES_RHM0_UVM90_LHM180_LVM270'].flatten().astype('float')
        behavior_angs = np.mod(90 - behavior_angs + 180, 360) - 180
        result = {'raw': pp_raw, 'matrix': behavior_mtx,
                  'mean': behavior_mean, 'sem': behavior_sem,
                  'angles': behavior_angs}
        return pyr.pmap(result)
    @pimms.value
    def abrams2012_data(pseudo_path):
        '''
        Load and return the data from Abrams et al. (2012).
        '''
        import os, numpy as np, pyrsistent as pyr
        from scipy.io import loadmat
        filename = pseudo_path.local_path('supp', 'abrams_nizam_carrasco_2012.mat')
        pp_raw = loadmat(filename)
        pp_dat = pp_raw['DATA']
        pp_dat = {nm: pp_dat[nm][0,0].T for nm in pp_dat.dtype.names}
        behavior_mtx = pp_dat['CS_75THRESH']
        behavior_mean = np.mean(behavior_mtx, axis=1)
        behavior_sem = np.sqrt(np.var(behavior_mtx, axis=1) / behavior_mtx.shape[1])
        behavior_angs = pp_dat['ANGLES'].flatten().astype('float')
        behavior_angs = np.mod(90 - behavior_angs + 180, 360) - 180
        result = {'raw': pp_raw, 'matrix': behavior_mtx,
                  'mean': behavior_mean, 'sem': behavior_sem,
                  'angles': behavior_angs}
        return pyr.pmap(result)

    # Tables for Plotting or General Summary #######################################################
    @staticmethod
    def _cortex_V1_mask(hem, skey):
        import numpy as np
        ii = hem.indices
        v1_sectors = tuple([lbl for (k,lbl) in skey.items()
                            if k[0] == 1 if k[3] >= 1 if k[4] <= 6])
        v1_sectors = np.array(v1_sectors)
        return np.isin(hem.prop('vpf_sector'), v1_sectors)
    @staticmethod
    def _subject_V1_surface_area(subjects, skey, sid, h, area_property='midgray_surface_area'):
        import numpy as np
        hem = subjects[sid].hemis[h]
        ii = VisualPerformanceFieldsDataset._cortex_V1_mask(hem, skey)
        return np.nansum(hem.property(area_property, mask=ii))
    @staticmethod
    def _subject_V1_thickness(subjects, skey, sid, h):
        import numpy as np
        hem = subjects[sid].hemis[h]
        ii = VisualPerformanceFieldsDataset._cortex_V1_mask(hem, skey)
        return np.nanmean(hem.property('thickness', mask=ii))
    @staticmethod
    def _subject_V1_volume(subjects, skey, sid, h, area_property='midgray_surface_area'):
        import numpy as np
        hem = subjects[sid].hemis[h]
        ii = VisualPerformanceFieldsDataset._cortex_V1_mask(hem, skey)
        area = hem.property(area_property, mask=ii)
        thic = hem.property('thickness', mask=ii)
        return np.nansum(area * thic)
    @staticmethod
    def _subject_cortex_surface_area(subjects, sid, h, area_property='midgray_surface_area'):
        import numpy as np
        hem = subjects[sid].hemis[h]
        ii = hem.mask('cortex_label')
        return np.nansum(hem.property(area_property, mask=ii))
    @staticmethod
    def _subject_cortex_volume(subjects, sid, h, area_property='midgray_surface_area'):
        import numpy as np
        hem = subjects[sid].hemis[h]
        ii = hem.mask('cortex_label')
        area = hem.property(area_property, mask=ii)
        thic = hem.property('thickness', mask=ii)
        return np.nansum(area * thic)
    @staticmethod 
    def _generate_restricted_v1_summary_call(tup):
        try:
            import neuropythy as ny, numpy as np
            data = ny.data['visual_performance_fields']
            subjects = data.subjects
            skey = data.sector_key[0]
            CLS = VisualPerformanceFieldsDataset
            (sid, saprop) = tup
            res = []
            for h in ['lh','rh']:
                v1sa = CLS._subject_V1_surface_area(subjects, skey, sid, h, area_property=saprop)
                v1vo = CLS._subject_V1_volume(subjects, skey, sid, h, area_property=saprop)
                cxsa = CLS._subject_cortex_surface_area(subjects, sid, h, area_property=saprop)
                cxvo = CLS._subject_cortex_volume(subjects, sid, h, area_property=saprop)
                v1th = CLS._subject_V1_thickness(subjects, skey, sid, h)
                res.append((v1sa, v1vo, cxsa, cxvo, v1th))
            return tuple(res)
        except Exception as e:
            import numpy as np, sys, os
            exc_type, exc_obj, exc_tb = sys.exc_info()
            fname = os.path.split(exc_tb.tb_frame.f_code.co_filename)[1]
            return ((np.nan, fname, exc_tb.tb_lineno, str(type(e)), str(e.args[0])),
                    (np.nan, fname, exc_tb.tb_lineno, str(type(e)), str(e.args[0])))
    @staticmethod
    def _generate_restricted_v1_summary_table(subjects, gender, agegroup,
                                              surface_area='midgray_surface_area',
                                              printstatus=False, nprocs=None):
        '''
        Generates and returns the summary table of V1 measurements regarding the
        1-6 degree eccentricity ROI. The resulting dataframe is formatted in a
        way that is deliberately friendly to seaborn. The resulting summary
        includes HCP restricted data.
        '''
        import neuropythy as ny, os, six, pyrsistent as pyr, multiprocessing as mp, numpy as np
        subject_list = subjects.keys()
        nsubs = len(subject_list)
        f = VisualPerformanceFieldsDataset._generate_restricted_v1_summary_call
        if nprocs is None or nprocs is Ellipsis: nprocs = mp.cpu_count()
        if nprocs < 2: nprocs = 1
        dat = ny.auto_dict(None, [])
        if printstatus:
            print('V1 Summary Processing blocks:')
        for ii in np.arange(0, nsubs, nprocs):
            mx = np.min([nsubs, ii + nprocs])
            sids = subject_list[ii:mx]
            if printstatus:
                print("  * %2d - %3d%% progress" % ((ii + 1)/nsubs * 100, mx/nsubs * 100))
            with mp.Pool(nprocs) as pool:
                sids = subject_list[ii:ii+nprocs]
                tups = pool.map(f, [(sid,surface_area) for sid in sids])
            # add these data into the overall roi table
            for (sid,hemtups) in zip(sids,tups):
                for (h, tup) in zip(['lh','rh'], hemtups):
                    dat['sid'].append(sid)
                    dat['Gender'].append(gender[sid])
                    dat['Age'].append(agegroup[sid])
                    dat['Hemisphere'].append(h)
                    (v1sa, v1vo, cxsa, cxvo, v1th) = tup
                    if np.isnan(v1sa):
                        raise ValueError('Subject %s, hemi %s, raised nan' % (sid, h),
                                         cxvo, v1th, v1vo, cxsa)
                    dat[r'V1 Surface Area [cm$^2$]'].append(v1sa / 100.0)
                    dat[r'V1 Gray Volume [cm$^3$]'].append(v1vo / 1000.0)
                    dat[r'Mean V1 Gray Thickness [mm]'].append(v1th)
                    dat[r'Cortex Surface Area [cm$^2$]'].append(cxsa / 100.0)
                    dat[r'Cortex Gray Volume [cm$^3$]'].append(cxvo / 1000.0)
                    dat[r'Normalized V1 Surface Area [%]'].append(v1sa / cxsa * 100)
                    dat[r'Normalized V1 Gray Volume [%]'].append(v1vo / cxvo * 100)
        return ny.to_dataframe(dat)
    def generate_restricted_v1_summary_table(self, surface_area='midgray_surface_area',
                                  printstatus=False, nprocs=None):
        '''
        Generates and returns the summary table of V1 measurements, including
        restricted HCP data, regarding the 1-6 degree eccentricity ROI. The
        resulting dataframe is formatted in a way that is deliberately friendly
        to seaborn.
        '''
        return type(self)._generate_v1_summary_table(self.subjects, self.gender, self.agegroup,
                                                     surface_area=surface_area,
                                                     printstatus=printstatus, nprocs=nprocs)
    @pimms.value
    def restricted_v1_summary_table(subjects, gender, agegroup):
        '''
        restricted_v1_pair_summary_table is a dataframe containing summary data,
        including restricted HCP data, about the V1 surface area, volume, and
        thickness for the ROIs examined in the associated paper (i.e., V1
        limited to 1-6 degrees of eccentricity). In order to access these data,
        you must have configured neuropythy to have access to the HCP restricted
        data.

        Note that because this table requires the restricted HCP data, it must be
        generated when requested (it cannot be stored on a public repository).
        Accordingly, it may take a substantial amount of time to load this
        value.
        '''
        return VisualPerformanceFieldsDataset._generate_restricted_v1_summary_table(subjects,
                                                                                    gender,
                                                                                    agegroup)
    @staticmethod 
    def _generate_v1_summary_call(tup):
        import neuropythy as ny, numpy as np
        data = ny.data['visual_performace_fields']
        subjects = data.subjects
        skey = data.sector_key[0]
        CLS = VisualPerformanceFieldsDataset
        (sid, saprop) = tup
        sub = data.subjects[sid]
        sas = []
        ths = []
        for h in ['lh','rh']:
            hem = sub.hemis[h]
            lbl = hem.prop('visual_area') == 1
            sas.append(hem.prop(saprop)[lbl])
            ths.append(hem.prop('thickness')[lbl])
        area = np.nansum(np.concatenate(sas))
        thick = np.nanmean(np.concatenate(ths))
        return (area, thick)
    @staticmethod
    def _generate_v1_summary_table(subjects, surface_area='midgray_surface_area',
                                   printstatus=False, nprocs=None):
        '''
        Generates and returns the summary table of V1 measurements for the
        entire V1 area as drawn by the raters for the HCP_lines dataset used in
        this project.
        '''
        import neuropythy as ny, os, six, pyrsistent as pyr, multiprocessing as mp, numpy as np
        subject_list = subjects.keys()
        nsubs = len(subject_list)
        f = VisualPerformanceFieldsDataset._generate_v1_summary_call
        if nprocs is None or nprocs is Ellipsis: nprocs = mp.cpu_count()
        if nprocs < 2: nprocs = 1
        dat = ny.auto_dict(None, [])
        if printstatus:
            print('V1 Summary Processing blocks:')
        for ii in np.arange(0, nsubs, nprocs):
            mx = np.min([nsubs, ii + nprocs])
            sids = subject_list[ii:mx]
            if printstatus:
                print("  * %2d - %3d%% progress" % ((ii + 1)/nsubs * 100, mx/nsubs * 100))
            with mp.Pool(nprocs) as pool:
                tups = pool.map(f, [(sid,surface_area) for sid in sids])
            # add these data into the overall roi table
            for (sid,tup) in zip(sids,tups):
                dat['sid'].append(sid)
                (sa, th) = tup
                dat['surface_area_cm2'].append(sa / 100.0)
                dat['mean_thickness_mm'].append(th)
        return ny.to_dataframe(dat)
    def generate_v1_summary_table(self, surface_area='midgray_surface_area',
                                  printstatus=False, nprocs=None):
        '''
        Generates and returns the summary table of V1 measurements regarding the
        full 0-6 degree eccentricity ROI.
        '''
        return type(self)._generate_v1_summary_table(self.subjects, surface_area=surface_area,
                                                     printstatus=printstatus, nprocs=nprocs)
    @pimms.value
    def v1_summary_table(subjects, pseudo_path):
        '''
        v1_summary_table is a dataframe containing summary data about the V1
        surface area, volume, and thickness for the ROIs examined in the
        associated paper (i.e., V1 limited to 0-6 degrees of eccentricity).
        '''
        import neuropythy as ny
        df = ny.load(pseudo_path.local_path('v1_summary_table.csv'))
        df.set_index(['sid'])
        return df
    @staticmethod
    def _generate_asymmetry_pair_table(subject_list, asymmetry, retinotopy_siblings,
                                       agegroup, gender,
                                       suffix='_cumulative'):
        import neuropythy as ny, pimms, numpy as np
        pair_data = ny.auto_dict(None, [])
        sid_to_index = {sid: k for (k,sid) in enumerate(np.sort(subject_list))}
        # Start by finding the subset of pairs we want to use:
        mzpairs = retinotopy_siblings['MZ']
        dzpairs = retinotopy_siblings['DZ']
        sbpairs = retinotopy_siblings['']
        # Find all the sibling pairs:
        sibs = set([tuple(sorted([k,v]))
                    for pairs in [mzpairs, dzpairs, sbpairs]
                    for (k,vs) in pairs.items()
                    for v in (vs if pimms.is_tuple(vs) else [vs])])
        # Now find unrelated pairs:
        urpairs = set([(a,b)
                       for a in subject_list
                       for b in subject_list
                       if a < b and (a,b) not in sibs])
        # The unrelated pairs get limited to age- and gender-matched pairs
        mzpairs = set([tuple(sorted([k,v]))
                       for (k,vs) in mzpairs.items()
                       for v in (vs if pimms.is_tuple(vs) else [vs])
                       if k in sid_to_index and v in sid_to_index])
        dzpairs = set([tuple(sorted([k,v]))
                       for (k,vs) in dzpairs.items()
                       for v in (vs if pimms.is_tuple(vs) else [vs])
                       if k in sid_to_index and v in sid_to_index])
        for (k,pairs) in [('MZ',mzpairs), ('DZ',dzpairs), ('UR',urpairs)]:
            for (t1,t2) in pairs:
                (k1,k2) = [sid_to_index.get(t, None) for t in (t1,t2)]
                if k1 is None or k2 is None: continue
                pair_data['sid_1'].append(t1)
                pair_data['sid_2'].append(t2)
                pair_data['relationship'].append(k)
                pair_data['same_age'].append(agegroup[t1] == agegroup[t2])
                pair_data['same_sex'].append(gender[t1] == gender[t2])
                for col in ['HVA','VMA']:
                    for (ii,w) in enumerate([10,20,30,40,50]):
                        pair_data['%s%d_1' % (col,w)].append(asymmetry[col+suffix][ii][k1])
                        pair_data['%s%d_2' % (col,w)].append(asymmetry[col+suffix][ii][k2])
        return ny.to_dataframe(pair_data)
    def generate_asymmetry_pair_table(self, suffix='_cumulative'):
        '''
        Yields a dataframe of asymmetry data sorted by pairs. Pair types include MZ (monozygotic
        twins), DZ (dizygotic twins), and 'UR' (unrelaited pair). This data will fail to load if you
        have not configured neuropythy to have access to the HCP restricted data.

        The optional argument suffix may be set to '' to calculate instantaneous instead of
        cumulative ROI asymmetry.
        '''
        return type(self)._generate_asymmetry_pair_table(self.subject_list, 
                                                         self.asymmetry, 
                                                         self.retinotopy_siblings, 
                                                         self.agegroup, 
                                                         self.gender)
    @pimms.value
    def asymmetry_pair_table(subject_list, asymmetry, retinotopy_siblings, agegroup, gender):
        '''
        asymmetry_pair_table is a dataframe of asymmetry data sorted by pairs. Pair types include
        MZ (monozygotic twins), DZ (dizygotic twins), and 'UR' (unrelaited pair). This data will
        fail to load if you have not configured neuropythy to have access to the HCP restricted
        data.
        '''
        return VisualPerformanceFieldsDataset._generate_asymmetry_pair_table(
            subject_list, asymmetry, retinotopy_siblings, agegroup, gender)
    
add_dataset('visual_performance_fields',
            lambda:VisualPerformanceFieldsDataset())



####################################################################################################
# neuropythy/plans/__init__.py
# A neuropythy module that stores PIMMS calculation plans.
# By Noah C. Benson

from .core import (limit_param, unlimit_param, imap_forget, imap_efferents)
from .prfclean import (prfclean_plan)

def run_prfclean(hemi, **kw):
    '''
    run_prfclean(hemi) runs the calculations associated with neuropythy.plans.prfclean_plan and
      yields a duplicate hemisphere object with the cleaned pRF maps included as properties. These
      properties

    In addition to including the cleaned pRF properties as properties of the returned hemisphere,
    this function appends meta-data describing the model parameters that were fit by the pRF
    cleaning.

    The following options may be given:
     * labels (one per vertex of coftex) or the name of a property of cortex.
       Note that the labels must not use 0 as an ROI: 0 must always indicate
       vertices that should not be optimized. Default: 'visual_area'
     * retinotopy: The argument for neuropythy's retinotopy_data function; this
       is typically a string prefix for retinotopy data on the cortex's
       properties ('prf_' for HCP subjects), or a map of the data itself.
       Default: 'prf_'.
     * tag: The prefix to give the output properties and the meta-data on the
       returned cortex object. The meta-data is named <teg>params.
       Default: 'prfclean_'
     * step_fn: A function that gets run every step of the optimization. The
       function must accept 2 arguments: the step number and the imap from the
       prfclean_plan. Default: None.
     * steps: The number of steps to run for minimization. Default: 200.
     * lr: The learning rate of the minimization. Default: 0.1.
     * max_eccen: The maximum eccentricity in the optimized maps. This should be
       the ecceentricity of the peripheral boundary of the labeled maps.
       Default: 90.
     * fieldsigns: The fieldsign values for each visual area. This may either be
       a dictionary whose keys are visual area labels and whose values are all
       either 1 or -1, or it can be Ellipsis or None. If the value is Ellipsis
       (the default), then the dictionary visual_area_field_signs from the
       neuropythy.vision package is used. If the value is None, then even
       labels are given fieldsigns of 1 and odd labels are given fieldsigns of
       -1.
     * model_knob: The base-2 log of the constant weight that the vmag_likelihood
       is multiplied by prior to summation with meas_likelihood. Default: 12.
     * boundary_knob: The base-2 log of the constant weight that the
       boundary_likelihood is multiplied by prior to summation with
       meas_likelihood. Default: 2.
     * start_coords: May specify the (2 x N) coordinate matrix of measurements
       to use as the starting point in the optimization. The N must be the
       number of vertices in the cortex; though values are only needed wherever
       there are non-zero labels. If None, then the measurements are used as
       the starting coordinates. The first row of the start coordinates must be
       the polar angle theta given in counter-clockwise radians starting from
       the right horizontal meridian, and the second row must be the
       eccentricity of the vertices in visual degrees.
     * prior_cmag_tan_params: The initial values of the tangential cortical
       magnificatiioin model parametters, whose interpretation depends on the
       value given for cmag_tan_method (see it for more details). Note that
       cmag_tan_method describes parameters such as loga, which are used in the
       model via the transformation a = exp(loga). The value that is given for
       the prior should *not* be the transformed param--i.e., provide a not loga
       as the prioir value.
     * prior_uvm_ipsi: The prior value of the upper-vertical-meridian's
       ipsilateral representation. This value is the number of radians into the
       ipsilateral visual field that the PRF representations of all visual areas
       extend for the upper vertical meridian.
     * prior_lvm_ipsi: The prior value of the lower-vertical-meridian's
       ipsilateral representation. This value is the number of radians into the
       ipsilateral visual field that the PRF representations of all visual areas
       extend for the lower vertical meridian.
     * grad_ipsi: Whether the uvm_ipsi and lvm_ipsi parameters, representing the
       ipsilateral representation in each visual area, are part of the optimized
       parameters whose gradients are required.
     * grad_coords: Specifies whether the tensors representing the coordinates of
       the PRF centers should track their gradients (default: True).
     * grad_cmag_tan: Specifies whether the tensors representing the tangential
       cortical magnification parameters of the retinotopy model should track
       their gradients (default: True).
     * grad_cmag_rad: Specifies whether the tensors representing the radial 
       cortical magnification parameters of the retinotopy model should track
       their gradients (default: True).
     * device: The pytorch device name to be used for all tensors (default: 'cpu').
     * meas_weight: A vector of weights, one per vertex in mesh, that specifies
       how strongly the model should believe that particular vertex's weight.
       Weights are the divisor for the meas_stddev. See calc_meas_stddev for
       more information about the standard deviation formula.
     * prior_cmag_eccen_offsets: The initial value of the offset parameter c2 
       from Horton and Hoyt's (1991) equation cmag(ecc) = c1 / (c2 + ecc).
       This should be either a single number, which is used for all visual
       areas, or a dictionary of visual area labels mapped to values.
     * dtype: May optionally specify the dtype for PyTorch to use in the
       optimization. This may be a PyTorch dtype object or it may be a string,
       in which case it is looked up as a member of the torch module (i.e.,
       'float' 
     * cmag_tan_method: The method to use for calculating tangential cortical
       magnification. This may be either 'sinusoid' (the default), 'beta', or
       None, in which case tangential cortical magnification is modeled to be
       uniform across polar angle. For thee sinusoid, the parameters are tan_hva
       and tan_vma where the actual hva and vma are arctan(tan_hva)*2/pi and
       arctan(tan_vma)*2/pi, limiting them both to the +/- 1 range. For the
       'beta' method, the params are loga and logb, where the beta-distribution
       parameters that are used are simply a = exp(loga) and b = exp(logb).
     * yield_imap: If True, yields the imap object instead of the hemi object
       after the minimization. Default: False.
    '''
    import numpy as np
    stepfn = kw.pop('step_fn', None)
    lr = kw.pop('lr', 0.1)
    steps = kw.pop('steps', 200)
    tag = kw.pop('tag', 'prfclean_')
    yield_imap = kw.pop('yield_imap', False)
    kw = dict(dict(retinotopy='prf_', labels='visual_area', max_eccen=90), **kw)
    # Go ahead and make the imap now:
    imap = prfclean_plan(cortex=hemi, **kw)
    llh0 = imap['likelihood']
    gradients = imap['gradients']
    grad_effs = imap_efferents(imap, gradients.keys())
    # Okay, we're going to run it!
    torch = imap['torch']
    opt = torch.optim.LBFGS(list(gradients.values()), lr=lr)
    for step in range(steps):
        if stepfn is not None:
            r = stepfn(step, imap)
            if r is not None: imap = r
        def closure():
            opt.zero_grad()
            imap_forget(imap, grad_effs)
            nll = -imap['likelihood']
            nll.backward()
            return nll
        opt.step(closure)
    # Finished minimizing--if we need to yield the imap we are done.
    if yield_imap: return imap
    # We need to put the data on the hemisphere.
    mesh = imap['mesh']
    ang = np.full(hemi.vertex_count, np.nan)
    ecc = np.full(hemi.vertex_count, np.nan)
    ang[mesh.labels] = np.mod(90 - 180/np.pi*imap['theta'].detach().numpy() + 180, 360) - 180
    ecc[mesh.labels] = imap['eccen'].detach().numpy()
    hemi = hemi.with_prop({(tag + 'polar_angle'): ang, (tag + 'eccentricity'): ecc})
    params = {k: np.array(imap[k].detach().numpy())
              for k in ['cmag_eccen_offsets', 'cmag_tan_params', 'uvm_ipsi', 'lvm_ipsi']}
    hemi = hemi.with_meta({(tag + 'params'): params})
    return hemi





####################################################################################################
# registration.py
# Tools for registering the cortical surface to a particular potential function
# By Noah C. Benson

'''
The neuropythy.registration package includes tools for performing cortical mesh registration to 2D
models that are projected to the cortical surface. See specifically, the help string for the 
mesh_register function.
'''

from .core       import (mesh_register, java_potential_term)



####################################################################################################
# neuropythy/util/filemap.py
# Utility for presenting a directory with a particular format as a data structure.
# By Noah C. Benson

import os, sys, warnings, logging, six, tarfile, atexit, shutil, posixpath, json, copy, pimms
import numpy          as np
import pyrsistent     as pyr
from   posixpath  import join as urljoin, split as urlsplit, normpath as urlnormpath
from   six.moves  import urllib
from   .core      import (library_path, curry, ObjectWithMetaData, AutoDict, data_struct, tmpdir,
                          is_tuple, is_list)
from   .conf      import to_credentials

# Not required, but try to load it anyway:
try:              import s3fs
except Exception: s3fs = None

def is_url(url):
    '''
    is_url(p) yields True if p is a valid URL and False otherwise.
    '''
    try: return bool(urllib.request.urlopen(url))
    except Exception: return False
def url_download(url, topath=None, create_dirs=True):
    '''
    url_download(url) yields the contents of the given url as a byte-string.
    url_download(url, topath) downloads the given url to the given path, topath and yields that path
      on success.

    The option create_dirs (default: True) may be set to False to prevent the topath directory from
    being created.
    '''
    # ensure directory exists
    if topath: topath = os.path.expanduser(os.path.expandvars(topath))
    if create_dirs and topath:
        dnm = os.path.dirname(topath)
        if not os.path.isdir(dnm): os.makedirs(os.path.abspath(dnm), 0o755)
    if six.PY2:
        response = urllib.request.urlopen(url)
        if topath is None: topath = response.read()
        else:
            with open(topath, 'wb') as fl:
                shutil.copyfileobj(response, fl)
    else:
        with urllib.request.urlopen(url) as response:
            if topath is None: topath = response.read()
            else:
                with open(topath, 'wb') as fl:
                    shutil.copyfileobj(response, fl)
    return topath
def is_s3_path(path):
    '''
    is_s3_path(path) yields True if path is a valid Amazon S3 path and False otherwise.
    '''
    return pimms.is_str(path) and path.lower().startswith('s3:')
def is_osf_path(path):
    '''
    is_osf_path(path) yields True if path starts with 'osf:' and False otherwise.
    '''
    return pimms.is_str(path) and path.lower().startswith('osf:')
tarball_endings = tuple([('.tar' + s) for s in ('','.gz','.bz2','.lzma')])
def split_tarball_path(path):
    '''
    split_tarball_path(path) yields a tuple (tarball, p) in which tarball is the path to the tarball
      referenced by path and p is the internal path that followed that tarball. If no tarball is
      included in the path, then (None, path) is returned. If no internal path is found following
      the tarball, then (path, '') is returned.
    '''
    lpath = path.lower()
    for e in tarball_endings:
        if lpath.endswith(e): return (path, '')
        ee = e + ':'
        if ee not in lpath: continue
        spl = path.split(ee)
        tarball = spl[0] + e
        p = ee.join(spl[1:])
        return (tarball, p)
    return (None, path)
def is_tarball_path(path):
    '''
    is_tarball_path(p) yields True if p is either a valid path of a tarball file or is such a path
      followed by a colon then any additional path.
    '''
    try: (tb,p) = split_tarball_path(path)
    except Exception: return False
    return tb is not None
def is_tarball_file(path):
    '''
    is_tarball_file(p) yields True if p is a valid path of a tarball file. Trailing :paths are not
      allowed (see is_tarball_path).
    '''
    (tb,p) = split_tarball_path(path)
    return tb is not None and p == ''
osf_basepath = 'https://api.osf.io/v2/nodes/%s/files/%s/'
def _osf_tree(proj, path=None, base='osfstorage'):
    if path is None: path = (osf_basepath % (proj, base))
    else:            path = (osf_basepath % (proj, base)) + path.lstrip('/')
    dat = json.loads(url_download(path, None))
    if 'data' not in dat: raise ValueError('Cannot detect kind of url for ' + path)
    res = {}
    if pimms.is_map(dat['data']): return dat['data']['links']['download']
    while dat is not None:
        links = dat.get('links', {})
        dat = dat['data']
        for u in dat:
            for r in [u['attributes']]:
                res[r['name']] = (u['links']['download'] if r['kind'] == 'file' else
                                  curry(lambda r: _osf_tree(proj, r, base), r['path']))
        nxt = links.get('next', None)
        if nxt is not None: dat = json.loads(url_download(nxt, None))
        else: dat = None
    return pimms.lazy_map(res)
def osf_crawl(k, *pths, **kw):
    '''
    osf_crawl(k) crawls the osf repository k and returns a lazy nested map structure of the
      repository's files. Folders have values that are maps of their contents while files have
      values that are their download links.
    osf_crawl(k1, k2, ...) is equivalent to osf_crawl(posixpath.join(k1, k2...)).

    The optional named argument base (default: 'osfstorage') may be specified to search in a
    non-standard storage position in the OSF URL; e.g. in the github storage.
    '''
    from six.moves import reduce
    base = kw.pop('base', 'osfstorage')
    root = kw.pop('root', None)
    if len(kw) > 0: raise ValueError('Unknown optional parameters: %s' % (list(kw.keys()),))
    if k.lower().startswith('osf:'): k = k[4:]
    k = k.lstrip('/')
    pths = [p.lstrip('/') for p in (k.split('/') + list(pths))]
    (bpth, pths) = (pths[0].strip('/'), [p for p in pths[1:] if p != ''])
    if root is None: root = _osf_tree(bpth, base=base)
    return reduce(lambda m,k: m[k], pths, root)

class BasicPath(object):
    '''
    BasicPath is the core path type that has utilities for handling various kinds of path
    operations through the pseudo-path interface.
    '''
    def __init__(self, base_path, pathmod=posixpath, cache_path=None):
        object.__setattr__(self, 'base_path', base_path)
        object.__setattr__(self, 'pathmod', pathmod)
        object.__setattr__(self, 'sep', pathmod.sep)
        object.__setattr__(self, 'cache_path', cache_path)
    def __setattr__(self, k, v): raise TypeError('path objects are immutable')
    def join(self, *args):
        args = [a for a in args if a != '']
        if len(args) == 0: return ''
        else: return self.pathmod.join(*args)
    def osjoin(self, *args):
        args = [a for a in args if a != '']
        if len(args) == 0: return ''
        else: return os.path.join(*args)
    def posixjoin(self, *args):
        args = [a for a in args if a != '']
        if len(args) == 0: return ''
        else: return posixpath.join(*args)
    def split(self, *args): return self.pathmod.split(*args)
    def to_ospath(self, path):
        path = self.pathmod.normpath(path)
        ps = []
        while path != '':
            (path,fl) = self.pathmod.split(path)
            ps.append(fl)
        return os.path.join(*reversed(ps))
    def base_find(self, rpath):
        '''
        Checks to see if the given path exists/can be found at the given base source; does not check
        the local cache and instead just checks the base source. If the file is not found, yields
        None; otherwise yields the relative path from the base_path.
        '''
        # this base version of the function assumes that pathmod.exists is sufficient:
        flnm = self.join(self.base_path, rpath)
        if self.pathmod.exists(flnm): return rpath
        else: return None
    def ensure_path(self, rpath, cpath):
        '''
        Called whenever a cache-path is supplied to getpath and the file is not already found there;
        must return a file path on the local filesystem.
        '''
        # for the base, we have no action here; for other types we may need to copy/download/extract
        # something then return that path.
        flnm = self.join(self.base_path, rpath)
        if os.path.exists(flnm): return flnm
        else: raise ValueError('ensure_path called for non-existant file: %s' % flnm)
    def _make_cache_path(self): return tmpdir(delete=True)
    def _cache_tarball(self, rpath, base_path=''):
        lpath = rpath[len(self.base_path):] if rpath.startswith(self.base_path) else rpath
        if self.cache_path is None: object.__setattr__(self, 'cache_path', self._make_cache_path())
        cpath = self.osjoin(self.cache_path, self.to_ospath(lpath))
        if os.path.isdir(cpath):
            (tbloc, tbinternal) = split_tarball_path(self.base_path)
            if tbloc is None: (tbloc, cpath) = (cpath, cpath)
            else:
                if self.base_path is None or self.base_path == '':
                    raise ValueError('attempting to cache unknown base file')
                tbname = self.split(tbloc)[-1]
                tbloc = cpath = self.osjoin(cpath, tbname)
                base_path = self.osjoin(base_path, tbinternal)
                if not os.path.exists(cpath): tbloc = self.ensure_path(rpath, cpath)
        else: tbloc = cpath
        tp = TARPath(tbloc, base_path, cache_path=cpath)
        if not os.path.exists(tp.tarball_path):
            self.ensure_path(lpath, tp.tarball_path)
        return tp
    def _check_tarball(self, *path_parts):
        rpath = self.join(*path_parts)
        # start by checking our base path:
        if self.base_path is not None and self.base_path != '':
            (tbloc, tbinternal) = split_tarball_path(self.base_path)
            if tbloc is not None:
                if tbinternal == '':
                    # we're fine; we just need to cache the current file...
                    return (self._cache_tarball(''), rpath)
                else:
                    # We copy ourselves to handle this base-path
                    tmp = copy.copy(self)
                    object.__setattr__(tmp, 'base_path', tbloc)
                    rpath = self.join(tbinternal, rpath)
                    # we defer to this path object with the new relative path:
                    return (tmp._cache_tarball(''), rpath)
        # okay, next check the relative path
        fpath = self.join('' if self.base_path is None else self.base_path, rpath)
        (tbloc, tbinternal) = split_tarball_path(fpath)
        if tbloc is not None:
            tbp = self._cache_tarball(tbloc)
            if tbinternal == '':
                return (None, tbp.tarball_path)
            else:
                return (tbp, tbinternal)
        # otherwise, we have no tarball on the path and just need to return ourselves as we are:
        return (self, rpath)
    def find(self, *path_parts):
        yes = self.join(*path_parts)
        (pp, rpath) = self._check_tarball(*path_parts)
        # if pp is none, we are requesting a cached tarball path rpath
        if pp is None: return yes
        # if that gave us back a different path object, we defer to it:
        if pp is not self: return yes if pp.find(rpath) is not None else None
        # check the cache path first
        if self.cache_path is not None:
            cpath = self.osjoin(self.cache_path, self.to_ospath(rpath))
            if os.path.exists(cpath): return yes
        # okay, check the base
        return self.base_find(rpath)
    def exists(self, *path_parts):
        return self.find(*path_parts) is not None
    def getpath(self, *path_parts):
        (pp, rpath) = self._check_tarball(*path_parts)
        # if pp is none, we are requesting a cached tarball path rpath
        if pp is None: return rpath
        # if that gave us back a different path object, we defer to it:
        if pp is not self: return pp.getpath(rpath)
        # check the cache path first
        rp = pp.find(rpath)
        fpath = self.join(self.base_path, rpath)
        if rp is None: raise ValueError('getpath: path not found: %s' % fpath)
        if self.cache_path is not None:
            cpath = self.osjoin(self.cache_path, self.to_ospath(rp))
            if len(rp) == 0:
                # we point to a file and are caching it locally...
                flnm = self.split(fpath)[-1]
                cpath = os.path.join(cpath, flnm)
            if os.path.exists(cpath): return cpath
            return self.ensure_path(rp, cpath)
        else: return self.ensure_path(rp, None)
    def ls(self, *path_parts):
        (pp, rpath) = self._check_tarball(*path_parts)
        # if pp is None, we are requesting a cached tarball path rpath
        if pp is None:
            # We want a listing of the tarball's root.
            pp = self._cache_tarball(rpath)
            rpath = ''
        # Whether pp is or is not this object, we defer to it.
        return pp.listdir(rpath)
    def listdir(self, rpath):
        raise TypeError("type %s does not implement listdir" % (type(self).__name__,))

class OSPath(BasicPath):
    def __init__(self, base_path, cache_path=None):
        BasicPath.__init__(self, base_path, pathmod=os.path, cache_path=cache_path)
    def to_ospath(self, path): return path
    def listdir(self, path):
        return os.listdir(os.path.join(self.base_path, path))
class URLPath(BasicPath):
    def __init__(self, base_path, cache_path=None):
        BasicPath.__init__(self, base_path, pathmod=posixpath, cache_path=cache_path)
    def base_find(self, rpath):
        if is_url(self.join(self.base_path, rpath)): return rpath
        else: return None
    def ensure_path(self, rpath, cpath):
        url = self.join(self.base_path, rpath)
        cdir = os.path.split(cpath)[0]
        if not os.path.isdir(cdir): os.makedirs(cdir, mode=0o755)
        return url_download(url, cpath)
class OSFPath(BasicPath):
    def __init__(self, base_path, cache_path=None):
        BasicPath.__init__(self, base_path, pathmod=posixpath, cache_path=cache_path)
        # if there's a tarball on the base path, we need to grab it instead of doing a typical
        # osf_crawl
        object.__setattr__(self, 'osf_tree', Ellipsis)
    def _find_url(self, rpath):
        if self.osf_tree is Ellipsis:
            (tbloc, tbinternal) = split_tarball_path(self.base_path)
            if tbloc == None: tree = osf_crawl(self.base_path)
            else: tree = osf_crawl(tbloc)
            object.__setattr__(self, 'osf_tree', tree)
        fl = self.osf_tree
        parts = [s for s in rpath.split(self.sep) if s != '']
        # otherwise walk the tree...
        for pp in parts:
            if   pp == '':                           continue
            elif not pimms.is_str(fl) and pp in fl:  fl = fl[pp]
            else:                                    return None
        return fl
    def base_find(self, rpath):
        fl = self._find_url(rpath)
        if fl is None: return None
        else: return rpath
    def ensure_path(self, rpath, cpath):
        fl = self._find_url(rpath)
        if not pimms.is_str(fl):
            if not os.path.isdir(cpath): os.makedirs(cpath, mode=0o755)
            return cpath
        else:
            cdir = os.path.split(cpath)[0]
            if not os.path.isdir(cdir): os.makedirs(cdir, mode=0o755)
        return url_download(fl, cpath)
    def listdir(self, path):
        fl = self._find_url(path)
        if fl is None: raise FileNotFoundError(path)
        if pimms.is_str(fl): raise NotADirectoryError(path)
        return list(fl.keys())
class S3Path(BasicPath):
    def __init__(self, fs, base_path, cache_path=None):
        BasicPath.__init__(self, base_path, pathmod=posixpath, cache_path=cache_path)
        object.__setattr__(self, 's3fs', fs)
    def base_find(self, rpath):
        fpath = self.join(self.base_path, rpath)
        if self.s3fs.exists(fpath): return rpath
        else: return None
    def ensure_path(self, rpath, cpath):
        url = self.join(self.base_path, rpath)
        cdir = os.path.split(cpath)[0]
        if not os.path.isdir(cdir): os.makedirs(cdir, mode=0o755)
        self.s3fs.get(url, cpath)
        return cpath
    def listdir(self, path):
        fpath = self.join(self.base_path, path)
        if fpath.lower().startswith('s3://'): fpath = fpath[5:]
        fpath = fpath.strip('/')
        lst = self.s3fs.ls(fpath)
        if len(lst) == 0: raise FileNotFoundError(path)
        elif len(lst) == 1 and lst[0] == fpath: raise NotADirectoryError(path)
        n = len(fpath) + 1
        return [fl[n:] for fl in lst]
class TARPath(OSPath):
    def __init__(self, tarpath, basepath='', cache_path=None, tarball_name=None):
        OSPath.__init__(self, basepath, cache_path=cache_path)
        object.__setattr__(self, 'tarball_path', tarpath)
        object.__setattr__(self, 'base_path', basepath)
        if cache_path == tarpath:
            # we need to prep the path
            tarfl = os.path.split(tarpath)[1]
            cpath = os.path.join(cache_path, 'contents')
            tarpath = os.path.join(cache_path, tarfl)
            if os.path.isfile(cache_path):
                # we need to move things around
                td = tmpdir(delete=True)
                tmpfl = os.path.join(td,tarfl)
                shutil.move(cache_path, tmpfl)
                if not os.path.isdir(cpath): os.makedirs(cpath, mode=0o755)
                shutil.move(tmpfl, tarpath)
            object.__setattr__(self, 'tarball_path', tarpath)
            object.__setattr__(self, 'cache_path', cpath)
        # get the tarball 'name'
        flnm = os.path.split(self.tarball_path if tarball_name is None else tarball_name)[-1]
        tarball_name = flnm.split('.tar')[0]
        object.__setattr__(self, 'tarball_name', tarball_name)
    tarball_fileobjs = {}
    @staticmethod
    def tarball_fileobj(path):
        if path not in TARPath.tarball_fileobjs:
            TARPath.tarball_fileobjs[path] = tarfile.open(path, 'r')
        return TARPath.tarball_fileobjs[path]
    def base_find(self, rpath):
        rpath = self.join(self.base_path, rpath)
        # check the cache path for both this rpath and this rpath + tarball name:
        cpath = self.join(self.cache_path, rpath)
        if os.path.exists(cpath): return rpath
        rpalt = self.join(self.tarball_name, rpath)
        cpalt = self.join(self.cache_path, rpalt)
        if os.path.exists(cpath): return rpalt
        # okay, see if they're int he tarfile
        tfl = TARPath.tarball_fileobj(self.tarball_path)
        try: found = bool(tfl.getmember(rpath))
        except Exception: found = False # might be that we need to prepend the tarball-name path
        if found: return rpath
        try: found = bool(tfl.getmember(rpalt))
        except Exception: pass
        if found: return rpalt
        # could still have a ./ ...
        rpath = self.join('.', rpath)
        rpalt = self.join('.', rpalt)
        try: found = bool(tfl.getmember(rpath))
        except Exception: found = False # might be that we need to prepend the tarball-name path
        if found: return rpath
        try: found = bool(tfl.getmember(rpalt))
        except Exception: pass
        if found: return rpalt
        else: return None
    def ensure_path(self, rpath, cpath):
        # we ignore cpath in this case
        with tarfile.open(self.tarball_path, 'r') as tfl:
            tfl.extract(rpath, self.cache_path)
            return cpath
    def listdir(self, path):
        fpath = self.join(self.base_path, rpath)
        fpath = fpath.strip('/') + '/'
        fn = len(fpath)
        with tarfile.open(self.tarball_path, 'r') as tfl:
            allfiles = tfl.getmembers()
        mems = []
        isdir = None
        for fl in allfiles:
            if fl.name == fpath[:-1]:
                if fl.isdir(): isdir = True
                else: raise NotADirectoryError(fpath)
            elif fl.name.startswith(fpath):
                nm = fl.name[fn:]
                if len(nm.split('/')) > 1: continue
                mems.append(nm)
        if len(mems) == 0 and not isdir: raise FileNotFoundError(path)
        return mems

@pimms.immutable
class PseudoPath(ObjectWithMetaData):
    '''
    The PseudoPath class represents either directories themselves, tarballs, or URLs as if they were
    directories.
    '''
    def __init__(self, source_path, cache_path=None, delete=Ellipsis, credentials=None,
                 meta_data=None):
        ObjectWithMetaData.__init__(self, meta_data=meta_data)
        self.source_path = source_path
        self.cache_path = cache_path
        self.delete = delete
        self.credentials = credentials
    @pimms.param
    def source_path(sp):
        '''
        pseudo_path.source_path is the source path of the the given pseudo-path object.
        '''
        if sp is None: return os.path.join(os.path.sep)
        if is_tuple(sp) and isinstance(sp[0], s3fs.S3FileSystem): return sp
        if not pimms.is_str(sp): raise ValueError('source_path must be a string/path')
        if is_url(sp) or is_s3_path(sp): return sp
        return os.path.expanduser(os.path.expandvars(sp))
    @pimms.param
    def cache_path(cp):
        '''
        pseudo_path.cache_path is the optionally provided cache path; this is the same as the
        storage path unless this is None.
        '''
        if cp is None: return None
        if not pimms.is_str(cp): raise ValueError('cache_path must be a string')
        return os.path.expanduser(os.path.expandvars(cp))
    @pimms.param
    def delete(d):
        '''
        pseudo_path.delete is True if the pseudo_path self-deletes on Python exit and False
        otherwise; if this is Ellipsis, then self-deletes only when the cache-directory is created
        by the PseudoPath class and is a temporary directory (i.e., not explicitly provided).
        '''
        if d in (True, False, Ellipsis): return d
        else: raise ValueError('delete must be True, False, or Ellipsis')
    @pimms.param
    def credentials(c):
        '''
        pdir.credentials is a tuple (key, secret) for authentication with Amazon S3.
        '''
        if c is None: return None
        else: return to_credentials(c)
    @pimms.value
    def _path_data(source_path, cache_path, delete, credentials):
        need_cache = True
        rpr = None
        delete = True if delete is Ellipsis else delete
        cp = {'cp':lambda:(tmpdir(delete=delete) if cache_path is None else cache_path)}
        cp = pimms.lazy_map(cp)
        # Okay, it might be a directory, an Amazon S3 URL, a different URL, or a tarball
        if is_tuple(source_path):
            (el0,pp) = (source_path[0], source_path[1:])
            if s3fs is not None and isinstance(el0, s3fs.S3FileSystem):
                base_path = urljoin(*pp)
                bpl = base_path.lower()
                if   bpl.startswith('s3://'): base_path = base_path[5:]
                elif bpl.startswith('s3:/'):  base_path = base_path[4:]
                elif bpl.startswith('s3:'):   base_path = base_path[3:]
                rpr = 's3://' + base_path
                pathmod = S3Path(el0, base_path, cache_path=cp['cp'])
            elif is_pseudo_path(el0):
                raise NotImplementedError('pseudo-path nests not yet supported')
            else: raise ValueError('source_path tuples must start with S3FileSystem or PseudoPath')
        elif os.path.isfile(source_path) and is_tarball_file(source_path):
            base_path = ''
            pathmod = TARPath(source_path, base_path, cache_path=cp['cp'])
            rpr = os.path.normpath(source_path)
        elif os.path.exists(source_path):
            rpr = os.path.normpath(source_path)
            base_path = source_path
            pathmod = OSPath(source_path, cache_path=cache_path)
        elif is_s3_path(source_path):
            if s3fs is None: raise ValueError('s3fs module is not installed')
            elif credentials is None: fs = s3fs.S3FileSystem(anon=True)
            else: fs = s3fs.S3FileSystem(key=credentials[0], secret=credentials[1])
            base_path = source_path
            pathmod = S3Path(fs, base_path, cache_path=cp['cp'])
            rpr = source_path
        elif is_osf_path(source_path):
            base_path = source_path
            pathmod = OSFPath(base_path, cache_path=cp['cp'])
            rpr = source_path
        elif is_url(source_path):
            base_path = source_path
            pathmod = URLPath(base_path, cache_path=cp['cp'])
            rpr = source_path
        elif is_tarball_path(source_path):
            # must be a a file starting with a tarball:
            (tbloc, tbinternal) = split_tarball_path(source_path)
            pathmod = TARPath(tbloc, tbinternal, cache_path=cp['cp'])
            rpr = os.path.normpath(source_path)
            base_path = ''
            # ok, don't know what it is...
        else: raise ValueError('Could not interpret source path: %s' % source_path)
        tmp = {'repr':rpr, 'pathmod':pathmod}
        if not cp.is_lazy('cp'):
            tmp['cache_path'] = cp['cp']
        return pyr.pmap(tmp)
    @pimms.require
    def check_path_data(_path_data):
        '''
        Ensures that _path_data is created without error.
        '''
        return ('pathmod' in _path_data)
    @pimms.value
    def actual_cache_path(_path_data):
        '''
        pdir.actual_cache_path is the cache path being used by the pseudo-path pdir; this may differ
          from the pdir.cache_path if the cache_path provided was None yet a temporary cache path
          was needed.
        '''
        return _path_data.get('cache_path', None)
    @pimms.value
    def actual_source_path(source_path):
        '''
        pdir.actual_source_path is identical to pdir.source_path except when the input source_path
        is a tuple (e.g. giving an s3fs object followed by a source path), in which case
        pdir.actual_source_path is a string representation of the source path.
        '''
        if source_path is None: return None
        if pimms.is_str(source_path): return source_path
        s = urljoin(*source_path[1:])
        if not s.lower().startswith('s3://'): s = 's3://' + s
        return s
    def __repr__(self):
        p = self._path_data['repr']
        return "pseudo_path('%s')" % p
    def join(self, *args):
        '''
        pdir.join(args...) is equivalent to os.path.join(args...) but always appropriate for the
          kind of path represented by the pseudo-path pdir.
        '''
        join = self._path_data['pathmod'].join
        return join(*args)
    def find(self, *args):
        '''
        pdir.find(paths...) is similar to to os.path.join(paths...) but it only yields the joined
          relative path if it can be found inside pdir; otherwise None is yielded. Note that this
          does not extract or download the path--it merely ensures that it exists.
        '''
        pmod = self._path_data['pathmod']
        return pmod.find(*args)
    def listpath(self, *args):
        '''
        pdir.listpath(paths...) returns a list of the files that are part of the subdirectory
           given by joining the paths. The pseudo-path pdir must support directory listing.
        '''
        pmod = self._path_data['pathmod']
        return pmod.ls(*args)
    def local_path(self, *args):
        '''
        pdir.local_path(paths...) is similar to os.path.join(pdir, paths...) except that it
          additionally ensures that the path being requested is found in the pseudo-path pdir then
          ensures that this path can be found in a local directory by downloading or extracting it
          if necessary. The local path is yielded.
        '''
        pmod = self._path_data['pathmod']
        return pmod.getpath(*args)
    def local_cache_path(self, *args):
        '''
        pdir.local_cache_path(paths...) is similar to os.path.join(pdir, paths...) except that it
          yields a local version of the given path, much like pdir.local_path(paths...). The 
          local_cache_path function differs from the local_path function in that, if no existing
          file is found at the given destination, no error is raised and the path is still returned.
        '''
        # if the file exists in the pseudo-path, just return the local path
        if self.find(*args) is not None: return self.local_path(*args)
        cp = self._path_data.get('cache_path', None)
        if cp is None: cp = self.source_path
        return os.path.join(cp, *args)
    def subpath(self, *args):
        '''
        pdir.subpath(paths...) returns a new pseudo-path object that points to the subdirectory
          implied by the given paths, which are joined. The new pseudo-path uses the matching
          subdirectory of pdir for its cache-path.
        '''
        cp = self.actual_cache_path
        sp = self.source_path
        if cp is not None:
            cp = os.path.join(cp, *args)
        sp = self.join(sp, *args)
        return PseudoPath(sp, cache_path=cp, delete=False, credentials=self.credentials,
                          meta_data={'superpath': self})
def is_pseudo_path(pdir):
    '''
    is_pseudo_path(pdir) yields True if the given object pdir is a pseudo-path object and False
      otherwise.
    '''
    return isinstance(pdir, PseudoPath)
def pseudo_path(source_path, cache_path=None, delete=Ellipsis, credentials=None, meta_data=None):
    '''
    pseudo_path(source_path) yields a pseudo-pathectory object that represents files in the given
      source path.

    pseudo-path objects act as an interface for loading data from abstract sources. The given source
    path may be either a directory, a (possibly zipped) tarball, or a URL. In all cases but the
    local directory, the pseudo-path object will quietly extract/download the requested files to a
    cache directory as their paths are requested. This is managed through two methods:
      * find(args...) joins the argument list as in os.path.join, then, if the resulting file is
        found in the source_path, this (relative) path-name is returned; otherwise None is returned.
      * local_path(args...) joins the argument list as in os.path.join, then, if the resulting file
        is found in the source_path, it is extracted/downloaded to the local cache directory if
        necessary, and this path (or the original path when no cache directory is used) is returned.

    The following optional arguments may be given:
      * cache_path (default: None) specifies the cache directory in which to put any extracted or
        downloaded contents. If None, then a temporary directory is created and used. If the source
        path is a local directory, then the cache path is not needed and is instead ignored. Note
        that if the cache path is not deleted, it can be reused across sessions--the pseudo-path
        will always check for files in the cache path before extracting or downloading them.
      * delete (default: Ellipsis) may be set to True or False to declare that the cache directory
        should be deleted at system exit (assuming a normal Python system exit). If Ellipsis, then
        the cache_path is deleted only if it is created by the pseudo-path object--given cache paths
        are never deleted.
      * credentials (default: None) may be set to a valid set of Amazon S3 credentials for use if
        the source path is an S3 path. The contents are passed through the to_credentials function.
      * meta_data (default: None) specifies an optional map of meta-data for the pseudo-path.
    '''
    return PseudoPath(source_path, cache_path=cache_path, delete=delete, credentials=credentials,
                     meta_data=meta_data)
def to_pseudo_path(obj):
    '''
    to_pseudo_path(obj) yields a pseudo-path object that has been coerced from the given obj or
      raises an exception. If the obj is a pseudo-path already, it is returned unchanged.
    '''
    if   is_pseudo_path(obj):   return obj
    elif pimms.is_str(obj):    return pseudo_path(obj)
    elif pimms.is_vector(obj):
        if len(obj) > 0 and pimms.is_map(obj[-1]): (obj,kw) = (obj[:-1],obj[-1])
        else: kw = {}
        return pseudo_path(*obj, **kw)
    else: raise ValueError('cannot coerce given object to a pseudo-path: %s' % obj)
    
@pimms.immutable
class FileMap(ObjectWithMetaData):
    '''
    The FileMap class is a pimms immutable class that tracks a set of FileMap format instructions
    with a valid path containing data of that format.
    '''
    def __init__(self, path, instructions, path_parameters=None, data_hierarchy=None,
                 cache_path=None, cache_delete=Ellipsis, load_function=None, meta_data=None, **kw):
        ObjectWithMetaData.__init__(self, meta_data=meta_data)
        self.path = path
        self.instructions = instructions
        self.data_hierarchy = data_hierarchy
        self.supplemental_paths = kw
        self.path_parameters = path_parameters
        self.load_function = load_function
        self.cache_path = cache_path
        self.cache_delete = cache_delete
    _tarball_endings = tuple([('.tar' + s) for s in ('','.gz','.bz2','.lzma')])
    @staticmethod
    def valid_path(p):
        '''
        FileMap.valid_path(path) yields os.path.abspath(path) if path is either a directory or a
          tarball file; yields path if path is a URL or s3 path, and otherwise yields None. If the
          path is a tarball path with a trailing inner-path, then the abspath of the tarball with
          the inner path appended is yielded.
        '''
        if   os.path.isdir(p):   return os.path.abspath(p)
        elif is_s3_path(p):      return p
        elif is_url(p):          return p
        # could still be a tarball path
        (tb,p) = split_tarball_path(p)
        if   tb is None: return None
        elif p  == '':   return os.path.abspath(tb)
        else:            return os.path.abspath(tb) + ':' + p
    @pimms.param
    def load_function(lf):
        '''
        filemap.load_function is the function used to load data by the filemap. It must accept
        exactly two arguments: filename and filedata. The file-data object is a merged map of both
        the path_parameters, meta_data, and file instruction, left-to-right in that order.
        '''
        if lf is None:
            from ..io import load
            return lambda flnm, meta: load(flnm, meta['format']) if 'format' in meta else load(flnm)
        else: return lf
    @pimms.param
    def cache_path(cp):
        '''
        filemap.cache_path is the path into which the filemap's data are cached (if necessary). This
        parameter represents the provided cache_path argument, but may not be the actual used cache
        path, which can be found in the actual_cache_path member.
        '''
        if cp is None: return None
        cp = os.path.expanduser(os.path.expandvars(cp))
        if not os.path.isdir(cp): raise ValueError('cache_path must be a directory or None')
        return cp
    @pimms.param
    def cache_delete(d):
        '''
        filemap.cache_delete is the setting used to decide whether to delete the cache directory at
        system exit. If Ellipsis (the default), then deletes any temporary directory created;
        otherwise does as this variable instructs.
        '''
        if d in [True,False,Ellipsis]: return d
        raise ValueError('cache_delete must be True, False, or Ellipsis')
    @pimms.param
    def path(p):
        '''
        filemap.path is the root path of the filemap object. 
        '''
        p0 = 0
        if is_pseudo_path(p): return p
        p = FileMap.valid_path(p)
        if p is None: raise ValueError('Path must be a directory or a tarball: %s' % p0)
        else: return p
    @pimms.param
    def instructions(inst):
        '''
        filemap.instructions is the map of load/save instructions for the given filemap.
        '''
        if not pimms.is_map(inst) and not isinstance(inst, list):
            raise ValueError('instructions must be a map or a list')
        return pimms.persist(inst)
    @pimms.param
    def data_hierarchy(h):
        '''
        filemap.data_hierarchy is the initial data hierarchy provided to the filemap object.
        '''
        return pimms.persist(h)
    @pimms.param
    def supplemental_paths(sp):
        '''
        filemap.supplemental_paths is a map of additional paths provided to the filemap object.
        '''
        if not pimms.is_map(sp): raise ValueError('supplemental_paths must be a map')
        rr = {}
        for (nm,pth) in six.iteritems(sp):
            pth = FileMap.valid_path(pth)
            if pth is None: raise ValueError('supplemental paths must be directories or tarballs')
            rr[nm] = pth
        return pimms.persist(rr)
    @pimms.param
    def path_parameters(pp):
        '''
        filemap.path_parameters is a map of parameters for the filemap's path.
        '''
        if pp is None: return pyr.m()
        elif not pimms.is_map(pp): raise ValueError('path perameters must be a mapping')
        else: return pimms.persist(pp)
    @staticmethod
    def parse_instructions(inst, hierarchy=None):
        '''
        FileMap.parse_instructions(inst) yields the tuple (data_files, data_tree); data_files is a
          map whose keys are relative filenames and whose values are the instruction data for the
          given file; data_tree is a lazy/nested map structure of the instruction data using 'type'
          as the first-level keys.

        The optional argument hierarchy specifies the hierarchy of the data to return in the
        data_tree. For example, if hierarchy is ['hemi', 'surface', 'roi'] then a file with the
        instructions {'type':'property', 'hemi':'lh', 'surface':'white', 'roi':'V1', 'name':'x'}
        would appear at data_tree['hemi']['lh']['surface']['white']['roi']['V1']['property']['x']
        whereas if hierarchy were ['roi', 'hemi', 'surface'] it would appear at
        data_tree['roi']['V1']['surface']['white']['hemi']['lh']['property']['x']. By default the
        ordering is undefined.
        '''
        dirstack = []
        data_tree = {}
        data_files = {}
        hierarchies = hierarchy if hierarchy else []
        if len(hierarchies) > 0 and pimms.is_str(hierarchies[0]): hierarchies = [hierarchies]
        known_filekeys = ('load','filt','when','then','miss')
        hierarchies = list(hierarchies)
        def handle_file(inst):
            # If it's a tuple, we just do each of them
            if isinstance(inst, tuple):
                for ii in inst: handle_file(ii)
                return None
            # first, walk the hierarchies; if we find one that matches, we use it; otherwise we make
            # one up
            dat = None
            for hrow in hierarchies:
                if not all(h in inst for h in hrow): continue
                dat = data_tree
                for h in hrow:
                    v = inst[h]
                    if h not in dat: dat[h] = {}
                    dat = dat[h]
                    if v not in dat: dat[v] = {}
                    dat = dat[v]
                break
            if dat is None:
                # we're gonna make up a hierarchy
                hh = []
                dat = data_tree
                for (k,v) in six.iteritems(inst):
                    if k in known_filekeys: continue
                    hh.append(k)
                    if k not in dat: dat[k] = {}
                    dat = dat[k]
                    if v not in dat: dat[v] = {}
                    dat = dat[v]
                # append this new hierarchy to the hierarchies
                hierarchies.append(hh)
            # Okay, we have the data, get the filename
            flnm = os.path.join(*dirstack)
            # add this data ot the data tree
            dat[flnm] = pyr.pmap(inst).set('_relpath', flnm)
            data_files[flnm] = inst
            return None
        def handle_dir(inst):
            # iterate over members
            dnm = None
            for k in inst:
                if dnm: dnm = handle_inst(k, dnm)
                elif pimms.is_str(k): dnm = k
                elif not isinstance(k, tuple) or len(k) != 2: raise ValueError('Bad dir content')
                else: dnm = handle_inst(k, dnm)
            return None
        def handle_inst(inst, k=None):
            if k: dirstack.append(k)
            if pimms.is_map(inst): handle_file(inst)
            elif isinstance(inst, (list, tuple)):
                if len(inst) == 0 or not pimms.is_map(inst[0]): handle_dir(inst)
                else: handle_file(inst)
            else: raise ValueError('Illegal instruction type: %s' % (inst,))
            if k: dirstack.pop()
            return None
        handle_inst(inst)
        return (data_files, data_tree)
    @pimms.value
    def _parsed_instructions(instructions, data_hierarchy):
        return pimms.persist(FileMap.parse_instructions(instructions, data_hierarchy))
    @pimms.value
    def actual_path(path):
        '''
        filemap.actual_path is always a string path (even when filemap.path is a pseudo-path).
        '''
        if is_pseudo_path(path): return path.source_path
        else: return path
    @pimms.value
    def actual_cache_path(cache_path, cache_delete, actual_path, supplemental_paths):
        '''
        filemap.actual_cache_path is the cache path used by the filemap, if needed.
        '''
        path = actual_path
        if cache_path is not None: return cache_path
        if (is_url(path) or is_s3_path(path) or is_tarball_path(path) or
            any(is_url(s) or is_s3_path(s) or is_tarball_path(s)
                for s in six.itervalues(supplemental_paths))):
            # we need a cache path...
            return tmpdir(delete=(True if cache_delete is Ellipsis else cache_delete))
        else: return None
    @pimms.require
    def setup_cache_deletion(cache_path, actual_cache_path, cache_delete):
        '''
        The filemap.setup_cache_deletion requirement ensures that the cache directory will be
        deleted at system exit, if required.
        '''
        if cache_delete is True and cache_path is actual_cache_path:
            atexit.register(shutil.rmtree, cache_path)
        return True
    @staticmethod
    def _load(pdir, flnm, loadfn, *argmaps, **kwargs):
        inst = pimms.merge(*(argmaps + (kwargs,)))
        flnm = flnm.format(**inst)
        args = pimms.merge(*argmaps, **kwargs)
        #logging.info('FileMap: loading file "%s"...\n' % flnm) #debug
        try:
            lpth = pdir.local_path(flnm)
            #logging.info('     ... local path: %s\n' % lpth) #debug
            args = pimms.merge(*argmaps, **kwargs)
            loadfn = inst['load'] if 'load' in args else loadfn
            #filtfn = inst['filt'] if 'filt' in args else lambda x,y:x
            dat = loadfn(lpth, args)
            #dat = filtfn(dat, args)
        except Exception:
            dat = None
            #raise
        # check for miss instructions if needed
        if dat is None and 'miss' in argmaps: miss = args['miss']
        else: miss = None
        if pimms.is_str(miss) and miss.lower() in ('error','raise','exception'):
            raise ValueError('File %s failed to load' % flnm)
        elif miss is not None:
            dat = miss(flnm, args)
        return dat
    @staticmethod
    def _parse_path(flnm, spaths, path_parameters, inst):
        flnm = flnm.format(**pimms.merge(path_parameters, inst))
        p0 = None
        for k in six.iterkeys(spaths):
            if k is None: continue
            if flnm.startswith(k + ':'):
                (flnm, p0) = (flnm[(len(k)+1):], k)
                break
        return (p0, flnm)
    @pimms.value
    def pseudo_paths(path, supplemental_paths, actual_cache_path):
        '''
        fmap.pseudo_paths is a mapping of pseduo-dirs in the file-map fmap. The primary path's
        pseudo-path is mapped to the key None.
        '''
        # we need to make cache paths for some of these...
        spaths = {}
        cp = None
        n = 0
        for (s,p) in six.iteritems(supplemental_paths):
            if actual_cache_path:
                cp = os.path.join(actual_cache_path, 'supp', s)
                if not os.path.isdir(cp): os.makedirs(os.path.abspath(cp), 0o755)
                n += 1
            spaths[s] = pseudo_path(p, delete=False, cache_path=cp)
        if actual_cache_path:
            if n > 0: cp = os.path.join(actual_cache_path, 'main')
            else:     cp = actual_cache_path
            if not os.path.isdir(cp): os.makedirs(os.path.abspath(cp), 0o755)
        spaths[None] = (path if is_pseudo_path(path) else
                        pseudo_path(path,delete=False,cache_path=cp))
        return pyr.pmap(spaths)
    @pimms.value
    def data_files(pseudo_paths, path_parameters, load_function, meta_data, _parsed_instructions):
        '''
        filemap.data_files is a lazy map whose keys are filenames and whose values are the loaded
        files.
        '''
        (data_files, data_tree) = _parsed_instructions
        res = {}
        for (flnm, inst) in six.iteritems(data_files):
            (pathnm, fn) = FileMap._parse_path(flnm, pseudo_paths, path_parameters, inst)
            res[fn] = curry(FileMap._load,
                            pseudo_paths[pathnm], flnm, load_function,
                            path_parameters, meta_data, inst)
        return pimms.lazy_map(res)
    @pimms.value
    def data_tree(_parsed_instructions, pseudo_paths, supplemental_paths, path_parameters,
                  data_files):
        '''
        filemap.data_tree is a lazy data-structure of the data loaded by the filemap's instructions.
        '''
        data_tree = _parsed_instructions[1]
        def visit_data(d):
            d = {k:visit_maps(v) for (k,v) in six.iteritems(d)}
            return data_struct(d)
        def lookup(flnm, inst):
            val = data_files[flnm]
            if 'filt' in inst: val = inst['filt'](val)
            return val
        def visit_maps(m):
            r = {}
            anylazy = False
            for (k,v) in six.iteritems(m):
                kk = k if isinstance(k, tuple) else [k]
                for k in kk:
                    if len(v) > 0 and '_relpath' in next(six.itervalues(v)):
                        (flnm,inst) = next(six.iteritems(v))
                        flnm = FileMap._parse_path(flnm, pseudo_paths, path_parameters, inst)[1]
                        r[k] = curry(lookup, flnm, inst)
                        anylazy = True
                    else: r[k] = visit_data(v)
            return pimms.lazy_map(r) if anylazy else pyr.pmap(r)
        return visit_data(data_tree)
def is_file_map(fmap):
    '''
    if_file_map(fmap) yields True if the given object fmap is a file map object and False otherwise.
    '''
    return isinstance(fmap, FileMap)
def file_map(path, instructions, **kw):
    '''
    file_map(path, instructions) yields a file-map object for the given path and instruction-set.
    file_map(None, instructions) yields a lambda of exactly one argument that is equivalent to the
      following:  lambda p: file_map(p, instructions)
    
    File-map objects are pimms immutable objects that combine a format-spec for a directory 
    (instructions) with a directory to yield a lazily-loaded data object. The format-spec is not
    currently documented, but interested users should see the variable
    neuropythy.hcp.files.hcp_filemap_instructions.
    
    The following options can be given:
     * path_parameters (default: None) may be set to a map of parameters that are used to format the
       filenames in the instructions.
     * data_hierarchy (default: None) may specify how the data should be nested; see the variable
       neuropythy.hcp.files.hcp_filemap_data_hierarchy.
     * load_function (default: None) may specify the function that is used to load filenames; if
       None then neuropythy.io.load is used.
     * meta_data (default: None) may be passed on to the FileMap object.

    Any additional keyword arguments given to the file_map function will be used as supplemental
    paths.
    '''
    if path: return FileMap(path, instructions, **kw)
    else:    return lambda path:file_map(path, instructions, **kw)
